# ruff: noqa: ARG002, SLF001, S105, S106
"""Tests for ModelSelectionMixin.

Covers model setup, selection, computed capability checks,
and event dispatching on model change.
"""

from __future__ import annotations

from types import SimpleNamespace
from unittest.mock import MagicMock, patch

from appkit_assistant.backend.database.models import ThreadStatus
from appkit_assistant.backend.schemas import AIModel
from appkit_assistant.state.thread.model_selection import ModelSelectionMixin

_PATCH = "appkit_assistant.state.thread.model_selection"

# Access computed-var descriptors via __dict__ to avoid triggering
# the Reflex __get__ descriptor protocol (which needs inherited_vars).
_CV = ModelSelectionMixin.__dict__


def _ai_model(
    model_id: str = "gpt-4o",
    *,
    supports_tools: bool = True,
    supports_attachments: bool = True,
    supports_search: bool = True,
    supports_skills: bool = True,
    requires_role: str | None = None,
) -> AIModel:
    return AIModel(
        id=model_id,
        text=model_id,
        model=model_id,
        supports_tools=supports_tools,
        supports_attachments=supports_attachments,
        supports_search=supports_search,
        supports_skills=supports_skills,
        requires_role=requires_role,
    )


class _StubModelSelection(ModelSelectionMixin):
    """Stub providing expected state vars without Reflex runtime."""

    # Class-level attributes for type(self).xxx lookups in set_selected_model
    load_available_skills_for_user = MagicMock()
    persist_current_thread = MagicMock()

    def __init__(self) -> None:
        self.ai_models: list[AIModel] = []
        self.selected_model: str = ""
        self._thread = SimpleNamespace(state=ThreadStatus.NEW, ai_model="")
        self.web_search_enabled = False
        self._restore_mcp_selection = MagicMock()
        self._clear_uploaded_files = MagicMock()
        self._restore_skill_selection = MagicMock()
        self.modal_active_tab = "tools"
        self._current_user_id: int | None = None


def _make_state() -> _StubModelSelection:
    return _StubModelSelection()


# ============================================================================
# Computed vars
# ============================================================================


class TestComputedVars:
    def test_get_selected_model(self) -> None:
        state = _make_state()
        state.selected_model = "gpt-4o"
        assert _CV["get_selected_model"].fget(state) == "gpt-4o"

    def test_has_ai_models_empty(self) -> None:
        state = _make_state()
        assert _CV["has_ai_models"].fget(state) is False

    def test_has_ai_models_with_models(self) -> None:
        state = _make_state()
        state.ai_models = [_ai_model()]
        assert _CV["has_ai_models"].fget(state) is True

    def test_supports_tools_no_model(self) -> None:
        state = _make_state()
        state.selected_model = ""
        assert _CV["selected_model_supports_tools"].fget(state) is False

    def test_supports_tools_true(self) -> None:
        state = _make_state()
        state.selected_model = "gpt-4o"
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model(supports_tools=True)
        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = _CV["selected_model_supports_tools"].fget(state)
        assert result is True

    def test_supports_tools_false(self) -> None:
        state = _make_state()
        state.selected_model = "basic"
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model(supports_tools=False)
        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = _CV["selected_model_supports_tools"].fget(state)
        assert result is False

    def test_supports_tools_model_not_found(self) -> None:
        state = _make_state()
        state.selected_model = "unknown"
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = None
        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = _CV["selected_model_supports_tools"].fget(state)
        assert result is False

    def test_supports_attachments_no_model(self) -> None:
        state = _make_state()
        state.selected_model = ""
        assert _CV["selected_model_supports_attachments"].fget(state) is False

    def test_supports_attachments_true(self) -> None:
        state = _make_state()
        state.selected_model = "gpt-4o"
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model(supports_attachments=True)
        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = _CV["selected_model_supports_attachments"].fget(state)
        assert result is True

    def test_supports_search_no_model(self) -> None:
        state = _make_state()
        state.selected_model = ""
        assert _CV["selected_model_supports_search"].fget(state) is False

    def test_supports_search_true(self) -> None:
        state = _make_state()
        state.selected_model = "gpt-4o"
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model(supports_search=True)
        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = _CV["selected_model_supports_search"].fget(state)
        assert result is True

    def test_supports_skills_no_model(self) -> None:
        state = _make_state()
        state.selected_model = ""
        assert _CV["selected_model_supports_skills"].fget(state) is False

    def test_supports_skills_true(self) -> None:
        state = _make_state()
        state.selected_model = "gpt-4o"
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model(supports_skills=True)
        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = _CV["selected_model_supports_skills"].fget(state)
        assert result is True


# ============================================================================
# _setup_models
# ============================================================================


class TestSetupModels:
    def test_all_models_loaded(self) -> None:
        state = _make_state()
        models = [_ai_model("m1"), _ai_model("m2")]
        mock_mgr = MagicMock()
        mock_mgr.get_all_models.return_value = models
        mock_mgr.get_default_model.return_value = "m1"

        user = SimpleNamespace(roles=[])
        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state._setup_models(user)

        assert len(state.ai_models) == 2
        assert state.selected_model == "m1"

    def test_role_filtering(self) -> None:
        state = _make_state()
        models = [
            _ai_model("public"),
            _ai_model("premium", requires_role="admin"),
        ]
        mock_mgr = MagicMock()
        mock_mgr.get_all_models.return_value = models
        mock_mgr.get_default_model.return_value = "public"

        user = SimpleNamespace(roles=[])
        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state._setup_models(user)

        assert len(state.ai_models) == 1
        assert state.ai_models[0].id == "public"

    def test_role_matching(self) -> None:
        state = _make_state()
        models = [
            _ai_model("public"),
            _ai_model("premium", requires_role="admin"),
        ]
        mock_mgr = MagicMock()
        mock_mgr.get_all_models.return_value = models
        mock_mgr.get_default_model.return_value = "public"

        user = SimpleNamespace(roles=["admin"])
        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state._setup_models(user)

        assert len(state.ai_models) == 2

    def test_keeps_current_selection(self) -> None:
        state = _make_state()
        state.selected_model = "m2"
        models = [_ai_model("m1"), _ai_model("m2")]
        mock_mgr = MagicMock()
        mock_mgr.get_all_models.return_value = models
        mock_mgr.get_default_model.return_value = "m1"

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state._setup_models(SimpleNamespace(roles=[]))

        assert state.selected_model == "m2"

    def test_fallback_to_default(self) -> None:
        state = _make_state()
        state.selected_model = "removed"
        models = [_ai_model("m1"), _ai_model("m2")]
        mock_mgr = MagicMock()
        mock_mgr.get_all_models.return_value = models
        mock_mgr.get_default_model.return_value = "m1"

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state._setup_models(SimpleNamespace(roles=[]))

        assert state.selected_model == "m1"

    def test_fallback_to_first(self) -> None:
        state = _make_state()
        state.selected_model = "removed"
        models = [_ai_model("m1")]
        mock_mgr = MagicMock()
        mock_mgr.get_all_models.return_value = models
        mock_mgr.get_default_model.return_value = "nonexistent"

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state._setup_models(SimpleNamespace(roles=[]))

        assert state.selected_model == "m1"

    def test_no_models_available(self) -> None:
        state = _make_state()
        state.selected_model = "old"
        mock_mgr = MagicMock()
        mock_mgr.get_all_models.return_value = []
        mock_mgr.get_default_model.return_value = ""

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state._setup_models(SimpleNamespace(roles=[]))

        assert state.selected_model == ""
        assert state.ai_models == []

    def test_none_user(self) -> None:
        state = _make_state()
        models = [_ai_model("m1"), _ai_model("m2", requires_role="admin")]
        mock_mgr = MagicMock()
        mock_mgr.get_all_models.return_value = models
        mock_mgr.get_default_model.return_value = "m1"

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state._setup_models(None)

        # Only models without requires_role should be available
        assert len(state.ai_models) == 1


# ============================================================================
# set_selected_model
# ============================================================================


class TestSetSelectedModel:
    def test_basic_selection(self) -> None:
        state = _make_state()
        mock_mgr = MagicMock()
        mock_model = _ai_model(
            "gpt-4o",
            supports_search=True,
            supports_tools=True,
            supports_attachments=True,
            supports_skills=True,
        )
        mock_mgr.get_model.return_value = mock_model

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = state.set_selected_model("gpt-4o")

        assert state.selected_model == "gpt-4o"
        assert state._thread.ai_model == "gpt-4o"
        assert state.modal_active_tab == "tools"
        assert result is not None

    def test_model_not_found(self) -> None:
        state = _make_state()
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = None

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = state.set_selected_model("unknown")

        assert state.selected_model == "unknown"
        assert result is None

    def test_disables_search(self) -> None:
        state = _make_state()
        state.web_search_enabled = True
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model(supports_search=False)

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state.set_selected_model("basic")

        assert state.web_search_enabled is False

    def test_disables_tools(self) -> None:
        state = _make_state()
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model(supports_tools=False)

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state.set_selected_model("basic")

        state._restore_mcp_selection.assert_called_once_with([])

    def test_disables_attachments(self) -> None:
        state = _make_state()
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model(supports_attachments=False)

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state.set_selected_model("basic")

        state._clear_uploaded_files.assert_called_once()

    def test_disables_skills(self) -> None:
        state = _make_state()
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model(supports_skills=False)

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            state.set_selected_model("basic")

        state._restore_skill_selection.assert_called_once_with([])

    def test_persists_for_active_thread(self) -> None:
        state = _make_state()
        state._thread.state = ThreadStatus.ACTIVE
        state._current_user_id = 1
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model()

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = state.set_selected_model("gpt-4o")

        assert len(result) == 2  # load_skills + persist

    def test_no_persist_for_new_thread(self) -> None:
        state = _make_state()
        state._thread.state = ThreadStatus.NEW
        state._current_user_id = 1
        mock_mgr = MagicMock()
        mock_mgr.get_model.return_value = _ai_model()

        with patch(f"{_PATCH}.ModelManager", return_value=mock_mgr):
            result = state.set_selected_model("gpt-4o")

        assert len(result) == 1  # only load_skills



"""Table component for displaying and managing OpenAI skills."""

import reflex as rx

import appkit_mantine as mn
from appkit_assistant.backend.database.models import Skill
from appkit_assistant.components.skill_dialogs import (
    create_skill_modal,
    delete_skill_dialog,
)
from appkit_assistant.state.skill_admin_state import SkillAdminState
from appkit_ui.styles import sticky_header_style


def _skill_table_row(skill: Skill) -> rx.Component:
    """Render a single skill row."""
    return mn.table.tr(
        mn.table.td(
            mn.text(skill.name, size="sm", fw="500", style={"whiteSpace": "nowrap"}),
            min_width="200px",
        ),
        mn.table.td(
            mn.text(
                skill.description,
                size="sm",
                c="dimmed",
                line_clamp=2,
                title=skill.description,
            ),
        ),
        mn.table.td(
            mn.group(
                mn.select(
                    value=skill.required_role,
                    data=SkillAdminState.available_roles,
                    placeholder="nicht eingeschränkt",
                    size="xs",
                    clearable=True,
                    check_icon_position="right",
                    on_change=lambda val: SkillAdminState.update_skill_role(
                        skill.id, val
                    ),
                    w="160px",
                ),
                mn.box(
                    rx.cond(
                        SkillAdminState.updating_role_skill_id == skill.id,
                        rx.spinner(size="1"),
                    ),
                    width="16px",
                    display="flex",
                    align_items="center",
                    justify_content="center",
                    flex_shrink="0",
                ),
                align="center",
                gap="xs",
                wrap="nowrap",
            ),
            width="1%",
            style={"whiteSpace": "nowrap"},
        ),
        mn.table.td(
            mn.group(
                mn.switch(
                    checked=skill.active,
                    on_change=lambda checked: SkillAdminState.toggle_skill_active(
                        skill.id, checked
                    ),
                    size="sm",
                ),
                mn.box(
                    rx.cond(
                        SkillAdminState.updating_active_skill_id == skill.id,
                        rx.spinner(size="1"),
                    ),
                    width="16px",
                    display="flex",
                    align_items="center",
                    justify_content="center",
                    flex_shrink="0",
                ),
                align="center",
                gap="xs",
                wrap="nowrap",
            ),
            width="1%",
            style={"whiteSpace": "nowrap"},
        ),
        mn.table.td(
            rx.hstack(
                delete_skill_dialog(skill),
                spacing="2",
                align_items="center",
            ),
            width="1%",
            style={"whiteSpace": "nowrap"},
        ),
    )


def skills_table(
    role_labels: dict[str, str] | None = None,
    available_roles: list[dict[str, str]] | None = None,
    load_on_mount: bool = False,
) -> rx.Component:
    """Admin table for managing OpenAI skills.

    Args:
        role_labels: Mapping of role names to display labels.
        available_roles: List of available roles for filtering.
        load_on_mount: If True, load skills on component mount.
            Default is False (lazy load via admin page).
    """
    if role_labels is None:
        role_labels = {}
    if available_roles is None:
        available_roles = []

    return mn.stack(
        rx.flex(
            mn.button(
                "Neuen Skill anlegen",
                left_section=rx.icon("plus", size=16),
                size="sm",
                on_click=SkillAdminState.open_create_modal,
                disabled=~SkillAdminState.has_skill_models,
            ),
            mn.text_input(
                placeholder="Skills filtern...",
                left_section=rx.icon("search", size=16),
                left_section_pointer_events="none",
                value=SkillAdminState.search_filter,
                on_change=SkillAdminState.set_search_filter,
                size="sm",
                w="18rem",
            ),
            rx.spacer(),
            mn.select(
                data=SkillAdminState.skill_model_options,
                value=SkillAdminState.selected_model_id,
                on_change=SkillAdminState.set_selected_model,
                placeholder="Modell auswählen...",
                size="sm",
                w="14rem",
            ),
            mn.button(
                "Synchronisieren",
                left_section=rx.icon("refresh-cw", size=16),
                size="sm",
                variant="outline",
                on_click=SkillAdminState.sync_skills,
                loading=SkillAdminState.syncing,
                disabled=~SkillAdminState.has_skill_models,
            ),
            width="100%",
            margin_bottom="md",
            gap="12px",
            align="center",
        ),
        create_skill_modal(),
        rx.box(
            mn.loading_overlay(
                visible=SkillAdminState.loading | SkillAdminState.syncing,
                overlayProps={"min_height": "200px"},
            ),
            mn.table(
                mn.table.thead(
                    mn.table.tr(
                        mn.table.th(mn.text("Name", size="sm", fw="700")),
                        mn.table.th(mn.text("Beschreibung", size="sm", fw="700")),
                        mn.table.th(mn.text("Rolle", size="sm", fw="700")),
                        mn.table.th(mn.text("Aktiv", size="sm", fw="700")),
                        mn.table.th(mn.text("", size="sm")),
                        style=sticky_header_style,
                    ),
                ),
                mn.table.tbody(
                    rx.foreach(
                        SkillAdminState.filtered_skills,
                        _skill_table_row,
                    )
                ),
                sticky_header=True,
                sticky_header_offset="0px",
                striped=False,
                highlight_on_hover=True,
                highlight_on_hover_color=rx.color_mode_cond(
                    light="gray.0",
                    dark="dark.8",
                ),
                w="100%",
            ),
            position="relative",
            w="100%",
            min_height="200px",
        ),
        w="100%",
        on_mount=(
            lambda: (
                [
                    SkillAdminState.set_available_roles(available_roles, role_labels),
                    SkillAdminState.load_skill_models() if load_on_mount else None,
                    SkillAdminState.load_skills_with_toast() if load_on_mount else None,
                ]
                if load_on_mount
                else [SkillAdminState.set_available_roles(available_roles, role_labels)]
            )
        ),
    )



"""Tests for AuthErrorDetector."""

from appkit_assistant.backend.services.auth_error_detector import (
    AuthErrorDetector,
    get_auth_error_detector,
)


class TestAuthErrorDetector:
    """Test suite for AuthErrorDetector."""

    def test_is_auth_error_detects_401_status(self) -> None:
        """is_auth_error detects 401 in error message."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("Error 401: Unauthorized") is True
        assert detector.is_auth_error("HTTP 401") is True

    def test_is_auth_error_detects_403_status(self) -> None:
        """is_auth_error detects 403 in error message."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("Error 403: Forbidden") is True
        assert detector.is_auth_error("HTTP 403") is True

    def test_is_auth_error_detects_unauthorized_keyword(self) -> None:
        """is_auth_error detects 'unauthorized' keyword."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("Request unauthorized") is True
        assert detector.is_auth_error("Unauthorized access") is True
        assert detector.is_auth_error("UNAUTHORIZED") is True

    def test_is_auth_error_detects_forbidden_keyword(self) -> None:
        """is_auth_error detects 'forbidden' keyword."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("Access forbidden") is True
        assert detector.is_auth_error("Forbidden resource") is True

    def test_is_auth_error_detects_token_errors(self) -> None:
        """is_auth_error detects token-related errors."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("Invalid token provided") is True
        assert detector.is_auth_error("Token expired") is True
        assert detector.is_auth_error("authentication required") is True

    def test_is_auth_error_detects_access_denied(self) -> None:
        """is_auth_error detects 'access denied'."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("Access denied") is True
        assert detector.is_auth_error("access DENIED") is True

    def test_is_auth_error_detects_not_authenticated(self) -> None:
        """is_auth_error detects 'not authenticated'."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("User not authenticated") is True
        assert detector.is_auth_error("NOT AUTHENTICATED") is True

    def test_is_auth_error_detects_auth_required(self) -> None:
        """is_auth_error detects 'auth_required'."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("auth_required") is True
        assert detector.is_auth_error("AUTH_REQUIRED") is True

    def test_is_auth_error_returns_false_for_other_errors(self) -> None:
        """is_auth_error returns False for non-auth errors."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("Network timeout") is False
        assert detector.is_auth_error("500 Internal Server Error") is False
        assert detector.is_auth_error("Invalid request") is False
        assert detector.is_auth_error("Not found") is False

    def test_is_auth_error_case_insensitive(self) -> None:
        """is_auth_error is case-insensitive."""
        detector = AuthErrorDetector()

        assert detector.is_auth_error("UNAUTHORIZED") is True
        assert detector.is_auth_error("Unauthorized") is True
        assert detector.is_auth_error("unauthorized") is True

    def test_extract_error_text_from_dict(self) -> None:
        """extract_error_text extracts 'message' from dict."""
        detector = AuthErrorDetector()

        error = {"message": "Auth failed", "code": 401}
        text = detector.extract_error_text(error)

        assert text == "Auth failed"

    def test_extract_error_text_from_dict_fallback(self) -> None:
        """extract_error_text uses str() when no 'message' key."""
        detector = AuthErrorDetector()

        error = {"code": 401, "status": "error"}
        text = detector.extract_error_text(error)

        assert "401" in text

    def test_extract_error_text_from_object_with_message(self) -> None:
        """extract_error_text extracts 'message' attribute from object."""

        class MockError:
            message = "Authentication error"

        detector = AuthErrorDetector()
        text = detector.extract_error_text(MockError())

        assert text == "Authentication error"

    def test_extract_error_text_from_string(self) -> None:
        """extract_error_text returns string as-is."""
        detector = AuthErrorDetector()

        text = detector.extract_error_text("Simple error message")

        assert text == "Simple error message"

    def test_extract_error_text_from_exception(self) -> None:
        """extract_error_text converts Exception to string."""
        detector = AuthErrorDetector()

        error = ValueError("Invalid value")
        text = detector.extract_error_text(error)

        assert "Invalid value" in text

    def test_extract_error_text_from_none(self) -> None:
        """extract_error_text returns empty string for None."""
        detector = AuthErrorDetector()

        text = detector.extract_error_text(None)

        assert text == ""

    def test_find_matching_server_in_error_finds_server(self) -> None:
        """find_matching_server_in_error finds server by name in error."""

        class MockServer:
            def __init__(self, name: str):
                self.name = name

        detector = AuthErrorDetector()
        servers = [MockServer("GitHub"), MockServer("GitLab"), MockServer("Bitbucket")]

        matched = detector.find_matching_server_in_error(
            "Authentication failed for GitHub", servers
        )

        assert matched is not None
        assert matched.name == "GitHub"

    def test_find_matching_server_in_error_case_insensitive(self) -> None:
        """find_matching_server_in_error is case-insensitive."""

        class MockServer:
            def __init__(self, name: str):
                self.name = name

        detector = AuthErrorDetector()
        servers = [MockServer("GitHub")]

        matched = detector.find_matching_server_in_error("github auth error", servers)

        assert matched is not None
        assert matched.name == "GitHub"

    def test_find_matching_server_in_error_returns_none_when_not_found(self) -> None:
        """find_matching_server_in_error returns None when no match."""

        class MockServer:
            def __init__(self, name: str):
                self.name = name

        detector = AuthErrorDetector()
        servers = [MockServer("GitHub")]

        matched = detector.find_matching_server_in_error(
            "Authentication failed for AWS", servers
        )

        assert matched is None

    def test_find_matching_server_in_error_handles_empty_list(self) -> None:
        """find_matching_server_in_error handles empty server list."""
        detector = AuthErrorDetector()

        matched = detector.find_matching_server_in_error("Some error", [])

        assert matched is None

    def test_find_matching_server_in_error_returns_first_match(self) -> None:
        """find_matching_server_in_error returns first matching server."""

        class MockServer:
            def __init__(self, name: str):
                self.name = name

        detector = AuthErrorDetector()
        servers = [
            MockServer("Server A"),
            MockServer("Server B"),
            MockServer("Server A Clone"),
        ]

        matched = detector.find_matching_server_in_error("Error with Server A", servers)

        assert matched.name == "Server A"


class TestAuthErrorDetectorSingleton:
    """Test suite for singleton pattern."""

    def test_get_auth_error_detector_returns_instance(self) -> None:
        """get_auth_error_detector returns AuthErrorDetector."""
        detector = get_auth_error_detector()

        assert isinstance(detector, AuthErrorDetector)

    def test_get_auth_error_detector_returns_same_instance(self) -> None:
        """get_auth_error_detector returns singleton."""
        detector1 = get_auth_error_detector()
        detector2 = get_auth_error_detector()

        assert detector1 is detector2



from appkit_assistant.backend.processors.claude_responses_processor import (
    ClaudeResponsesProcessor,
)
from appkit_assistant.backend.processors.processor_base import ProcessorBase
from appkit_assistant.backend.processors.perplexity_processor import PerplexityProcessor
from appkit_assistant.backend.processors.streaming_base import StreamingProcessorBase
from appkit_assistant.backend.processors.openai_chat_completion_processor import (
    OpenAIChatCompletionsProcessor,
)
from appkit_assistant.backend.processors.openai_responses_processor import (
    OpenAIResponsesProcessor,
)
from appkit_assistant.backend.processors.lorem_ipsum_processor import (
    LoremIpsumProcessor,
)
from appkit_assistant.backend.processors.gemini_responses_processor import (
    GeminiResponsesProcessor,
)

__all__ = [
    "ClaudeResponsesProcessor",
    "GeminiResponsesProcessor",
    "LoremIpsumProcessor",
    "OpenAIChatCompletionsProcessor",
    "OpenAIResponsesProcessor",
    "PerplexityProcessor",
    "ProcessorBase",
    "StreamingProcessorBase",
]



"""Dialog components for AI model management."""

import logging

import reflex as rx

import appkit_mantine as mn
from appkit_assistant.backend.database.models import AssistantAIModel
from appkit_assistant.state.ai_model_admin_state import AIModelAdminState
from appkit_ui.components.dialogs import delete_dialog
from appkit_ui.components.form_inputs import form_field

logger = logging.getLogger(__name__)

_MAX_TEMPERATURE: float = 2.0

# Known processor types offered in the dropdown
KNOWN_PROCESSOR_TYPES = [
    {"value": "openai", "label": "OpenAI"},
    {"value": "claude", "label": "Anthropic Claude"},
    {"value": "perplexity", "label": "Perplexity"},
    {"value": "gemini", "label": "Google Gemini"},
    {"value": "lorem_ipsum", "label": "Lorem Ipsum (Test)"},
]

# Suggested icons for convenience - still a free-text field
KNOWN_ICONS = [
    {"value": "openai", "label": "OpenAI"},
    {"value": "anthropic", "label": "Anthropic"},
    {"value": "perplexity", "label": "Perplexity"},
    {"value": "googlegemini", "label": "Google Gemini"},
    {"value": "codesandbox", "label": "Codesandbox (Standard)"},
]


# Processor → default icon mapping
PROCESSOR_DEFAULT_ICONS: dict[str, str] = {
    "openai": "openai",
    "claude": "anthropic",
    "perplexity": "perplexity",
    "gemini": "googlegemini",
    "lorem_ipsum": "codesandbox",
}


class AIModelValidationState(rx.State):
    """Validation state for add/edit AI model forms."""

    model_id: str = ""
    text: str = ""
    icon: str = ""
    model: str = ""
    processor_type: str = ""
    temperature: str = "0.05"
    stream: bool = False
    supports_tools: bool = False
    supports_attachments: bool = False
    supports_search: bool = False
    supports_skills: bool = False
    requires_role: str = ""
    # API credentials (stored encrypted; leave blank to use global config)
    api_key: str = ""
    base_url: str = ""
    on_azure: bool = False

    model_id_error: str = ""
    text_error: str = ""
    processor_type_error: str = ""
    temperature_error: str = ""

    @rx.event
    def initialize(self, record: AssistantAIModel | None = None) -> None:
        """Reset validation state for add or edit mode."""
        if record is None:
            self.model_id = ""
            self.text = ""
            self.icon = ""
            self.model = ""
            self.processor_type = ""
            self.temperature = "0.05"
            self.stream = False
            self.supports_tools = False
            self.supports_attachments = False
            self.supports_search = False
            self.supports_skills = False
            self.requires_role = ""
            self.api_key = ""
            self.base_url = ""
            self.on_azure = False
        else:
            self.model_id = record.model_id
            self.text = record.text
            self.icon = record.icon or ""
            self.model = record.model or ""
            self.processor_type = record.processor_type
            self.temperature = str(record.temperature)
            self.stream = record.stream
            self.supports_tools = record.supports_tools
            self.supports_attachments = record.supports_attachments
            self.supports_search = record.supports_search
            self.supports_skills = record.supports_skills
            self.requires_role = record.requires_role or ""
            self.api_key = record.api_key or ""
            self.base_url = record.base_url or ""
            self.on_azure = record.on_azure

        self.model_id_error = ""
        self.text_error = ""
        self.processor_type_error = ""
        self.temperature_error = ""

    # --- Field setters with inline validation ---

    def set_model_id(self, value: str) -> None:
        self.model_id = value
        self.validate_model_id()

    def set_text(self, value: str) -> None:
        self.text = value
        self.validate_text()

    def set_icon(self, value: str) -> None:
        self.icon = value

    def set_model(self, value: str) -> None:
        self.model = value

    def set_processor_type(self, value: str) -> None:
        self.processor_type = value
        # Auto-set icon when it is still empty or the codesandbox default
        if not self.icon or self.icon == "codesandbox":
            self.icon = PROCESSOR_DEFAULT_ICONS.get(value, "codesandbox")
        self.validate_processor_type()

    def set_temperature(self, value: str) -> None:
        self.temperature = value
        self.validate_temperature()

    def set_stream(self, value: bool) -> None:
        self.stream = value

    def set_supports_tools(self, value: bool) -> None:
        self.supports_tools = value

    def set_supports_attachments(self, value: bool) -> None:
        self.supports_attachments = value

    def set_supports_search(self, value: bool) -> None:
        self.supports_search = value

    def set_supports_skills(self, value: bool) -> None:
        self.supports_skills = value

    def set_requires_role(self, value: str) -> None:
        self.requires_role = value

    def set_api_key(self, value: str) -> None:
        self.api_key = value

    def set_base_url(self, value: str) -> None:
        self.base_url = value

    def set_on_azure(self, value: bool) -> None:
        self.on_azure = value

    # --- Validators ---

    @rx.event
    def validate_model_id(self) -> None:
        if not self.model_id or not self.model_id.strip():
            self.model_id_error = "Modell-ID ist ein Pflichtfeld."
        else:
            self.model_id_error = ""

    @rx.event
    def validate_text(self) -> None:
        if not self.text or not self.text.strip():
            self.text_error = "Anzeigename ist ein Pflichtfeld."
        else:
            self.text_error = ""

    @rx.event
    def validate_processor_type(self) -> None:
        if not self.processor_type or not self.processor_type.strip():
            self.processor_type_error = "Prozessortyp ist ein Pflichtfeld."
        else:
            self.processor_type_error = ""

    @rx.event
    def validate_temperature(self) -> None:
        try:
            val = float(self.temperature)
            if val < 0 or val > _MAX_TEMPERATURE:
                self.temperature_error = "Temperatur muss zwischen 0 und 2 liegen."
            else:
                self.temperature_error = ""
        except ValueError:
            self.temperature_error = "Temperatur muss eine Zahl sein."

    @rx.var
    def has_errors(self) -> bool:
        return bool(
            self.model_id_error
            or self.text_error
            or self.processor_type_error
            or self.temperature_error
        )


def _role_select() -> rx.Component:
    """Role selection dropdown for model access control."""
    return mn.select(
        label="Erforderliche Rolle",
        description=("Nur Benutzer mit dieser Rolle können dieses Modell verwenden."),
        data=AIModelAdminState.available_roles,
        value=AIModelValidationState.requires_role,
        on_change=AIModelValidationState.set_requires_role,
        placeholder="Keine Einschränkung",
        clearable=True,
        name="requires_role",
        width="100%",
    )


def _processor_type_field() -> rx.Component:
    """Processor type select."""
    return mn.select(
        label="Prozessortyp",
        description="Welcher Prozessor verarbeitet Anfragen für dieses Modell?",
        data=KNOWN_PROCESSOR_TYPES,
        value=AIModelValidationState.processor_type,
        on_change=AIModelValidationState.set_processor_type,
        error=AIModelValidationState.processor_type_error,
        placeholder="Prozessortyp auswählen",
        required=True,
        allow_deselect=False,
        name="processor_type",
        width="100%",
    )


def _capability_switches() -> rx.Component:
    """Capability toggle switches for the model."""
    return mn.simple_grid(
        mn.switch(
            label="Streaming",
            checked=AIModelValidationState.stream,
            on_change=AIModelValidationState.set_stream,
            name="stream",
        ),
        mn.switch(
            label="Tools",
            checked=AIModelValidationState.supports_tools,
            on_change=AIModelValidationState.set_supports_tools,
            name="supports_tools",
        ),
        mn.switch(
            label="Attachments",
            checked=AIModelValidationState.supports_attachments,
            on_change=AIModelValidationState.set_supports_attachments,
            name="supports_attachments",
        ),
        mn.switch(
            label="Web Search",
            checked=AIModelValidationState.supports_search,
            on_change=AIModelValidationState.set_supports_search,
            name="supports_search",
        ),
        mn.switch(
            label="Skills",
            checked=AIModelValidationState.supports_skills,
            on_change=AIModelValidationState.set_supports_skills,
            name="supports_skills",
        ),
        cols=2,
        spacing="sm",
    )


def _modal_footer(
    submit_label: str,
    on_cancel: rx.EventHandler,
) -> rx.Component:
    """Footer buttons for add/edit modals."""
    return rx.flex(
        mn.button(
            "Abbrechen",
            variant="subtle",
            on_click=on_cancel,
        ),
        mn.button(
            submit_label,
            type="submit",
            disabled=AIModelValidationState.has_errors,
            loading=AIModelAdminState.loading,
        ),
        direction="row",
        gap="9px",
        justify_content="end",
        padding="16px",
        border_top="1px solid var(--mantine-color-default-border)",
        background="var(--mantine-color-body)",
        width="100%",
    )


def ai_model_form_fields() -> rx.Component:
    """Reusable form fields for add/edit modals."""
    return mn.flex(
        form_field(
            name="text",
            icon="tag",
            label="Anzeigename",
            hint="Anzeige-Label in der Modellauswahl",
            type="text",
            placeholder="GPT 5 Mini",
            default_value=AIModelValidationState.text,
            required=True,
            max_length=100,
            on_blur=AIModelValidationState.set_text,
            validation_error=AIModelValidationState.text_error,
        ),
        form_field(
            name="model_id",
            icon="hash",
            label="Modell-ID",
            hint="Eindeutige interne ID (z.B. gpt-5-mini)",
            type="text",
            placeholder="gpt-5-mini",
            default_value=AIModelValidationState.model_id,
            required=True,
            max_length=100,
            on_blur=AIModelValidationState.set_model_id,
            validation_error=AIModelValidationState.model_id_error,
        ),
        form_field(
            name="model",
            icon="box",
            label="API-Modellname",
            hint="Name des Modells in der API (z.B. gpt-5-mini)",
            type="text",
            placeholder="gpt-5-mini",
            default_value=AIModelValidationState.model,
            max_length=100,
            on_blur=AIModelValidationState.set_model,
        ),
        mn.select(
            label="Icon",
            description="Symbol für die Modellauswahl",
            data=KNOWN_ICONS,
            value=AIModelValidationState.icon,
            on_change=AIModelValidationState.set_icon,
            placeholder="Icon auswählen",
            searchable=True,
            clearable=True,
            name="icon",
            width="100%",
        ),
        _processor_type_field(),
        form_field(
            name="temperature",
            icon="thermometer",
            label="Temperatur",
            hint="Sampling-Temperatur (0.0 - 2.0)",
            type="number",
            placeholder="0.05",
            default_value=AIModelValidationState.temperature,
            on_blur=AIModelValidationState.set_temperature,
            validation_error=AIModelValidationState.temperature_error,
        ),
        mn.divider(label="Fähigkeiten", my="4px"),
        _capability_switches(),
        mn.divider(label="API-Zugangsdaten", my="4px"),
        mn.switch(
            label="Azure-Endpunkt",
            checked=AIModelValidationState.on_azure,
            on_change=AIModelValidationState.set_on_azure,
            name="on_azure",
        ),
        form_field(
            name="base_url",
            icon="link",
            label="Base URL",
            hint="Optionaler API-Endpunkt (leer = Standard)",
            type="text",
            placeholder="https://api.openai.com/v1",
            default_value=AIModelValidationState.base_url,
            max_length=500,
            on_blur=AIModelValidationState.set_base_url,
        ),
        mn.password_input(
            name="api_key",
            label="API-Schlüssel",
            description="Schlüssel für dieses Modell",
            placeholder="sk-...",
            default_value=AIModelValidationState.api_key,
            on_blur=AIModelValidationState.set_api_key,
            required=True,
            w="100%",
        ),
        mn.divider(label="Zugriffssteuerung", my="4px"),
        _role_select(),
        direction="column",
        gap="9px",
        padding="12px",
    )


def add_ai_model_button() -> rx.Component:
    """Button that opens the add AI model modal."""
    return mn.button(
        "Neues KI-Modell anlegen",
        left_section=rx.icon("plus", size=16),
        size="sm",
        on_click=[
            AIModelValidationState.initialize(record=None),
            AIModelAdminState.open_add_modal,
        ],
    )


def _ai_model_modal(
    title: str,
    opened: bool | rx.Var,
    on_close: rx.EventHandler,
    on_submit: rx.EventHandler,
    submit_label: str,
) -> rx.Component:
    """Shared modal structure for add/edit AI model."""
    return mn.modal(
        rx.form.root(
            rx.flex(
                mn.scroll_area.autosize(
                    ai_model_form_fields(),
                    max_height="60vh",
                    width="100%",
                    type="always",
                    offset_scrollbars=True,
                ),
                _modal_footer(submit_label, on_close),
                direction="column",
            ),
            on_submit=on_submit,
            reset_on_submit=False,
            height="100%",
        ),
        title=title,
        opened=opened,
        on_close=on_close,
        size="lg",
        centered=True,
        overlay_props={"backgroundOpacity": 0.5, "blur": 4},
    )


def add_ai_model_modal() -> rx.Component:
    """Modal for adding a new AI model."""
    return _ai_model_modal(
        title="Neues KI-Modell anlegen",
        opened=AIModelAdminState.add_modal_open,
        on_close=AIModelAdminState.close_add_modal,
        on_submit=AIModelAdminState.add_model,
        submit_label="KI-Modell anlegen",
    )


def edit_ai_model_modal() -> rx.Component:
    """Modal for editing an existing AI model."""
    return _ai_model_modal(
        title="KI-Modell aktualisieren",
        opened=AIModelAdminState.edit_modal_open,
        on_close=AIModelAdminState.close_edit_modal,
        on_submit=AIModelAdminState.modify_model,
        submit_label="KI-Modell aktualisieren",
    )


def delete_ai_model_dialog(record: AssistantAIModel) -> rx.Component:
    """Confirmation dialog for deleting an AI model."""
    return delete_dialog(
        title="KI-Modell löschen",
        content=record.text,
        on_click=lambda: AIModelAdminState.delete_model(record.id),
        icon_button=True,
        variant="subtle",
        color="red",
    )



"""
Gemini responses processor for generating AI responses using Google's GenAI API.
"""

import asyncio
import copy
import logging
import uuid
from collections.abc import AsyncGenerator
from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass, field
from typing import Any, Final, NamedTuple

import httpx
from google import genai
from google.genai import types
from mcp import ClientSession
from mcp.client.streamable_http import streamable_http_client

from appkit_assistant.backend.database.models import (
    MCPServer,
)
from appkit_assistant.backend.processors.mcp_mixin import MCPCapabilities
from appkit_assistant.backend.processors.processor_base import mcp_oauth_redirect_uri
from appkit_assistant.backend.processors.streaming_base import StreamingProcessorBase
from appkit_assistant.backend.schemas import (
    AIModel,
    Chunk,
    MCPAuthType,
    Message,
    MessageType,
)
from appkit_assistant.backend.services.system_prompt_builder import SystemPromptBuilder

logger = logging.getLogger(__name__)
default_oauth_redirect_uri: Final[str] = mcp_oauth_redirect_uri()

# Maximum characters to show in tool result preview
TOOL_RESULT_PREVIEW_LENGTH: Final[int] = 500

GEMINI_FORBIDDEN_SCHEMA_FIELDS: Final[set[str]] = {
    "$schema",
    "$id",
    "$ref",
    "$defs",
    "definitions",
    "$comment",
    "examples",
    "default",
    "const",
    "contentMediaType",
    "contentEncoding",
    "additionalProperties",
    "additional_properties",
    "patternProperties",
    "unevaluatedProperties",
    "unevaluatedItems",
    "minItems",
    "maxItems",
    "minLength",
    "maxLength",
    "minimum",
    "maximum",
    "exclusiveMinimum",
    "exclusiveMaximum",
    "multipleOf",
    "pattern",
    "format",
    "title",
    "allOf",
    "oneOf",
    "not",
    "if",
    "then",
    "else",
    "dependentSchemas",
    "dependentRequired",
}


@dataclass
class MCPToolContext:
    """Context for MCP tool execution."""

    session: ClientSession
    server_name: str
    tools: dict[str, Any] = field(default_factory=dict)


class MCPSessionWrapper(NamedTuple):
    """Wrapper to store MCP connection details before creating actual session."""

    url: str
    headers: dict[str, str]
    name: str


class GeminiResponsesProcessor(StreamingProcessorBase, MCPCapabilities):
    """Gemini processor using the GenAI API with native MCP support."""

    def __init__(
        self,
        models: dict[str, AIModel],
        api_key: str | None = None,
        oauth_redirect_uri: str = default_oauth_redirect_uri,
    ) -> None:
        StreamingProcessorBase.__init__(self, models, "gemini_responses")
        MCPCapabilities.__init__(self, oauth_redirect_uri, "gemini_responses")
        self.client: genai.Client | None = None
        self._system_prompt_builder = SystemPromptBuilder()

        if api_key:
            try:
                self.client = genai.Client(
                    api_key=api_key, http_options={"api_version": "v1beta"}
                )
            except Exception as e:
                logger.error("Failed to initialize Gemini client: %s", e)
        else:
            logger.warning("Gemini API key not found. Processor disabled.")

        logger.debug("Using redirect URI for MCP OAuth: %s", oauth_redirect_uri)

    def _get_event_handlers(self) -> dict[str, Any]:
        """Get event handlers (empty for Gemini - uses chunk-based handling)."""
        return {}

    async def process(
        self,
        messages: list[Message],
        model_id: str,
        files: list[str] | None = None,  # noqa: ARG002
        mcp_servers: list[MCPServer] | None = None,
        payload: dict[str, Any] | None = None,
        user_id: int | None = None,
        cancellation_token: asyncio.Event | None = None,
    ) -> AsyncGenerator[Chunk, None]:
        """Process messages using Google GenAI API with native MCP support."""
        if not self.client:
            raise ValueError("Gemini Client not initialized.")

        if model_id not in self.models:
            msg = f"Model {model_id} not supported by Gemini processor"
            raise ValueError(msg)

        model = self.models[model_id]
        self.current_user_id = user_id
        self.clear_pending_auth_servers()

        # Initialize statistics
        self._reset_statistics(model.model)

        # Prepare configuration
        config = self._create_generation_config(model, payload)

        # Connect to MCP servers and create sessions
        mcp_sessions = []
        mcp_prompt = ""
        if mcp_servers:
            mcp_sessions, auth_required = await self._create_mcp_sessions(
                mcp_servers, user_id
            )
            for server in auth_required:
                self.add_pending_auth_server(server)
            mcp_prompt = self._build_mcp_prompt(mcp_servers)
            # Note: tools are configured in _stream_with_mcp after connecting to MCP

        # Prepare messages with MCP prompts for tool selection
        contents, system_instruction = await self._convert_messages_to_gemini_format(
            messages, mcp_prompt
        )

        # Add system instruction to config if present
        if system_instruction:
            config.system_instruction = system_instruction

        if mcp_sessions:
            logger.info(
                "Connected to %d MCP servers for native tool support",
                len(mcp_sessions),
            )

        try:
            # Generate content with MCP tools
            async for chunk in self._stream_with_mcp(
                model.model, contents, config, mcp_sessions, cancellation_token
            ):
                yield chunk

            # Yield final completion chunk with statistics
            stats = self._get_statistics()
            logger.debug("Completion statistics: %s", stats)
            yield self.chunk_factory.completion(
                status="response_complete",
                statistics=stats,
            )

            # Handle any pending auth
            async for auth_chunk in self.yield_pending_auth_chunks():
                yield auth_chunk

        except Exception as e:
            logger.exception("Error in Gemini processor: %s", str(e))
            yield self.chunk_factory.error(f"Error: {e!s}")

    async def _create_mcp_sessions(
        self, servers: list[MCPServer], user_id: int | None
    ) -> tuple[list[MCPSessionWrapper], list[MCPServer]]:
        """Create MCP ClientSession objects for each server.

        Returns:
            Tuple with (sessions, auth_required_servers)
        """
        sessions = []
        auth_required = []

        for server in servers:
            try:
                # Parse headers using MCPCapabilities
                headers = self.parse_mcp_headers(server)

                # Handle OAuth - inject token
                if (
                    server.auth_type == MCPAuthType.OAUTH_DISCOVERY
                    and user_id is not None
                ):
                    token = await self.get_valid_token(server, user_id)
                    if token:
                        headers["Authorization"] = f"Bearer {token.access_token}"
                    else:
                        auth_required.append(server)
                        logger.warning(
                            "Skipping MCP server %s - OAuth token required",
                            server.name,
                        )
                        continue

                # Create SSE client connection
                # Use URL directly as configured (server determines endpoint)
                logger.debug(
                    "Connecting to MCP server %s at %s (headers: %s)",
                    server.name,
                    server.url,
                    {
                        k: "***" if k.lower() == "authorization" else v
                        for k, v in headers.items()
                    },
                )

                # Inject x-user-id header if enabled for this server
                if server.inject_user_id and self.current_user_id is not None:
                    headers["x-user-id"] = str(self.current_user_id)

                # Create a session wrapper with URL and headers
                session = MCPSessionWrapper(server.url, headers, server.name)
                sessions.append(session)

            except Exception as e:
                logger.error(
                    "Failed to connect to MCP server %s: %s", server.name, str(e)
                )

        return sessions, auth_required

    async def _stream_with_mcp(
        self,
        model_name: str,
        contents: list[types.Content],
        config: types.GenerateContentConfig,
        mcp_sessions: list[Any],
        cancellation_token: asyncio.Event | None = None,
    ) -> AsyncGenerator[Chunk, None]:
        """Stream responses with MCP tool support."""
        if not mcp_sessions:
            # No MCP sessions, direct streaming
            async for chunk in self._stream_generation(
                model_name, contents, config, cancellation_token
            ):
                yield chunk
            return

        # Enter all session contexts and fetch tools
        async with self._mcp_context_manager(mcp_sessions) as tool_contexts:
            if tool_contexts:
                # Convert MCP tools to Gemini FunctionDeclarations
                # Use unique naming: server_name__tool_name to avoid duplicates
                function_declarations = []
                for ctx in tool_contexts:
                    for tool_name, tool_def in ctx.tools.items():
                        # Create unique name: server_name__tool_name
                        unique_name = f"{ctx.server_name}__{tool_name}"
                        func_decl = self._mcp_tool_to_gemini_function(
                            unique_name, tool_def, original_name=tool_name
                        )
                        if func_decl:
                            function_declarations.append(func_decl)

                if function_declarations:
                    config.tools = [
                        types.Tool(function_declarations=function_declarations)
                    ]
                    logger.info(
                        "Configured %d tools for Gemini from MCP",
                        len(function_declarations),
                    )

            # Stream with automatic function calling loop
            async for chunk in self._stream_with_tool_loop(
                model_name, contents, config, tool_contexts, cancellation_token
            ):
                yield chunk

    async def _stream_with_tool_loop(
        self,
        model_name: str,
        contents: list[types.Content],
        config: types.GenerateContentConfig,
        tool_contexts: list[MCPToolContext],
        cancellation_token: asyncio.Event | None = None,
    ) -> AsyncGenerator[Chunk, None]:
        """Stream generation with tool call handling loop."""
        max_tool_rounds = 10
        current_contents = list(contents)

        for _round_num in range(max_tool_rounds):
            if cancellation_token and cancellation_token.is_set():
                logger.info("Processing cancelled by user")
                break

            # Use streaming API
            stream = await self.client.aio.models.generate_content_stream(
                model=model_name, contents=current_contents, config=config
            )

            # Collect function calls and stream text as it arrives
            collected_function_calls: list[Any] = []
            collected_parts: list[types.Part] = []
            streamed_text = ""

            async for chunk in stream:
                if cancellation_token and cancellation_token.is_set():
                    logger.info("Processing cancelled by user")
                    return

                if chunk.usage_metadata:
                    self._update_statistics(
                        input_tokens=chunk.usage_metadata.prompt_token_count,
                        output_tokens=chunk.usage_metadata.candidates_token_count,
                    )

                if not chunk.candidates or not chunk.candidates[0].content:
                    continue

                for part in chunk.candidates[0].content.parts:
                    # Stream text immediately
                    if part.text:
                        streamed_text += part.text
                        yield self.chunk_factory.text(part.text, delta=part.text)
                    # Collect function calls
                    if part.function_call is not None:
                        collected_function_calls.append(part.function_call)
                        collected_parts.append(part)

            # If no function calls, we're done (text was already streamed)
            if not collected_function_calls:
                return

            # Build the model's response content with function calls
            model_response_parts = []
            if streamed_text:
                model_response_parts.append(types.Part(text=streamed_text))
            model_response_parts.extend(collected_parts)
            current_contents.append(
                types.Content(role="model", parts=model_response_parts)
            )

            # Execute tool calls and collect results
            function_responses = []
            for fc in collected_function_calls:
                # Parse unique tool name: server_name__tool_name
                server_name, original_tool_name = self._parse_unique_tool_name(fc.name)

                # Generate a unique tool call ID
                tool_call_id = f"mcp_{uuid.uuid4().hex[:32]}"

                # Yield TOOL_CALL chunk to show in UI (use original name)
                yield self.chunk_factory.tool_call(
                    f"Benutze Werkzeug: {server_name}.{original_tool_name}",
                    tool_name=original_tool_name,
                    tool_id=tool_call_id,
                    server_label=server_name,
                    status="starting",
                )

                result = await self._execute_mcp_tool(fc.name, fc.args, tool_contexts)

                # Yield TOOL_RESULT chunk with preview
                preview = (
                    result[:TOOL_RESULT_PREVIEW_LENGTH]
                    if len(result) > TOOL_RESULT_PREVIEW_LENGTH
                    else result
                )
                yield self.chunk_factory.tool_result(
                    preview,
                    tool_id=tool_call_id,
                    status="completed",
                )

                function_responses.append(
                    types.Part(
                        function_response=types.FunctionResponse(
                            name=fc.name,
                            response={"result": result},
                        )
                    )
                )
                logger.debug(
                    "Tool %s executed, result length: %d",
                    fc.name,
                    len(str(result)),
                )

            # Add function responses
            current_contents.append(
                types.Content(role="user", parts=function_responses)
            )

            # Continue to next round

        logger.warning("Max tool rounds (%d) exceeded", max_tool_rounds)

    def _parse_unique_tool_name(self, unique_name: str) -> tuple[str, str]:
        """Parse unique tool name back to server name and original tool name.

        Args:
            unique_name: Tool name in format 'server_name__tool_name'

        Returns:
            Tuple of (server_name, original_tool_name)
        """
        if "__" in unique_name:
            parts = unique_name.split("__", 1)
            return parts[0], parts[1]
        # Fallback for tools without prefix
        return "unknown", unique_name

    async def _execute_mcp_tool(
        self,
        unique_tool_name: str,
        args: dict[str, Any],
        tool_contexts: list[MCPToolContext],
    ) -> str:
        """Execute an MCP tool and return the result."""
        server_name, tool_name = self._parse_unique_tool_name(unique_tool_name)

        for ctx in tool_contexts:
            if ctx.server_name == server_name and tool_name in ctx.tools:
                try:
                    logger.debug(
                        "Executing tool %s on server %s with args: %s",
                        tool_name,
                        server_name,
                        args,
                    )
                    result = await ctx.session.call_tool(tool_name, args)
                    if hasattr(result, "content") and result.content:
                        texts = [i.text for i in result.content if hasattr(i, "text")]
                        return "\n".join(texts) if texts else str(result)
                    return str(result)
                except Exception as e:
                    logger.exception("Error executing tool %s: %s", tool_name, e)
                    return f"Error executing tool: {e!s}"

        return f"Tool {tool_name} not found on server {server_name}"

    def _mcp_tool_to_gemini_function(
        self,
        name: str,
        tool_def: dict[str, Any],
        original_name: str | None = None,
    ) -> types.FunctionDeclaration | None:
        """Convert MCP tool definition to Gemini FunctionDeclaration.

        Args:
            name: Unique function name (may include server prefix)
            tool_def: MCP tool definition with description and inputSchema
            original_name: Original tool name for description enhancement
        """
        try:
            description = tool_def.get("description", "")
            # Enhance description with original name if using prefixed naming
            if original_name and original_name != name:
                description = f"[{original_name}] {description}"
            input_schema = tool_def.get("inputSchema", {})

            # Fix the schema for Gemini compatibility
            fixed_schema = self._fix_schema_for_gemini(input_schema)

            return types.FunctionDeclaration(
                name=name,
                description=description,
                parameters=fixed_schema if fixed_schema else None,
            )
        except Exception as e:
            logger.warning("Failed to convert MCP tool %s: %s", name, str(e))
            return None

    def _fix_schema_for_gemini(self, schema: dict[str, Any]) -> dict[str, Any]:
        """Fix JSON schema for Gemini API compatibility (recursive)."""
        if not isinstance(schema, dict):
            return schema

        result = copy.deepcopy(schema)

        # Remove forbidden fields
        for key in GEMINI_FORBIDDEN_SCHEMA_FIELDS:
            result.pop(key, None)

        # Fix array without items
        if result.get("type") == "array" and "items" not in result:
            result["items"] = {"type": "string"}

        # Recurse into nested schemas
        if isinstance(result.get("items"), dict):
            result["items"] = self._fix_schema_for_gemini(result["items"])

        if isinstance(result.get("properties"), dict):
            result["properties"] = {
                k: self._fix_schema_for_gemini(v)
                for k, v in result["properties"].items()
            }

        for key in ("anyOf", "any_of"):
            if isinstance(result.get(key), list):
                result[key] = [
                    self._fix_schema_for_gemini(item) for item in result[key]
                ]

        return result

    @asynccontextmanager
    async def _mcp_context_manager(
        self, session_wrappers: list[Any]
    ) -> AsyncGenerator[list[MCPToolContext], None]:
        """Context manager to enter all MCP session contexts and fetch tools."""
        async with AsyncExitStack() as stack:
            tool_contexts: list[MCPToolContext] = []

            for wrapper in session_wrappers:
                try:
                    logger.debug(
                        "Connecting to MCP server %s via streamablehttp_client",
                        wrapper.name,
                    )
                    # Create httpx client with headers and timeout
                    http_client = httpx.AsyncClient(
                        headers=wrapper.headers,
                        timeout=60.0,
                    )
                    # Register client for cleanup
                    await stack.enter_async_context(http_client)

                    read, write, _ = await stack.enter_async_context(
                        streamable_http_client(
                            url=wrapper.url,
                            http_client=http_client,
                        )
                    )

                    session = await stack.enter_async_context(
                        ClientSession(read, write)
                    )
                    await session.initialize()

                    # Fetch tools from this session
                    tools_result = await session.list_tools()
                    tools_dict = {}
                    for tool in tools_result.tools:
                        tools_dict[tool.name] = {
                            "description": tool.description or "",
                            "inputSchema": (
                                tool.inputSchema if hasattr(tool, "inputSchema") else {}
                            ),
                        }

                    ctx = MCPToolContext(
                        session=session,
                        server_name=wrapper.name,
                        tools=tools_dict,
                    )
                    tool_contexts.append(ctx)

                    logger.info(
                        "MCP session initialized for %s with %d tools",
                        wrapper.name,
                        len(tools_dict),
                    )
                except Exception as e:
                    logger.exception(
                        "Failed to initialize MCP session for %s: %s",
                        wrapper.name,
                        str(e),
                    )

            try:
                yield tool_contexts
            except Exception as e:
                logger.exception("Error during MCP session usage: %s", str(e))
                raise

    async def _stream_generation(
        self,
        model_name: str,
        contents: list[types.Content],
        config: types.GenerateContentConfig,
        cancellation_token: asyncio.Event | None = None,
    ) -> AsyncGenerator[Chunk, None]:
        """Stream generation from Gemini model."""
        # generate_content_stream returns an awaitable that yields an async generator
        stream = await self.client.aio.models.generate_content_stream(
            model=model_name, contents=contents, config=config
        )
        async for chunk in stream:
            if cancellation_token and cancellation_token.is_set():
                logger.info("Processing cancelled by user")
                break
            processed = self._handle_chunk(chunk)
            if processed:
                yield processed

    def _create_generation_config(
        self, model: AIModel, payload: dict[str, Any] | None
    ) -> types.GenerateContentConfig:
        """Create generation config from model and payload."""
        # Default thinking level depends on model
        thinking_level = "medium" if "flash" in model.model.lower() else "high"

        # Override from payload if present
        if payload:
            thinking_level = payload.get("thinking_level", thinking_level)

        # Filter out fields not accepted by GenerateContentConfig
        filtered_payload = {}
        if payload:
            ignored_fields = {"thread_uuid", "user_id", "thinking_level"}
            filtered_payload = {
                k: v for k, v in payload.items() if k not in ignored_fields
            }

        return types.GenerateContentConfig(
            temperature=model.temperature,
            thinking_config=types.ThinkingConfig(thinking_level=thinking_level),
            **filtered_payload,
            response_modalities=["TEXT"],
        )

    def _build_mcp_prompt(self, mcp_servers: list[MCPServer]) -> str:
        """Build MCP tool selection prompt from server prompts."""
        return "\n".join(f"- {s.prompt}" for s in mcp_servers if s.prompt)

    async def _convert_messages_to_gemini_format(
        self, messages: list[Message], mcp_prompt: str = ""
    ) -> tuple[list[types.Content], str | None]:
        """Convert app messages to Gemini Content objects."""
        system_instruction = await self._system_prompt_builder.build(mcp_prompt)
        contents: list[types.Content] = []

        for msg in messages:
            match msg.type:
                case MessageType.SYSTEM:
                    system_instruction = (
                        f"{system_instruction}\n{msg.text}"
                        if system_instruction
                        else msg.text
                    )
                case MessageType.HUMAN | MessageType.ASSISTANT:
                    role = "user" if msg.type == MessageType.HUMAN else "model"
                    contents.append(
                        types.Content(role=role, parts=[types.Part(text=msg.text)])
                    )

        return contents, system_instruction

    def _extract_text_from_parts(self, parts: list[Any]) -> str:
        """Extract and join text from content parts."""
        return "".join(p.text for p in parts if p.text)

    def _handle_chunk(self, chunk: Any) -> Chunk | None:
        """Handle a single chunk from Gemini stream."""
        # Update statistics if usage metadata is present in the chunk
        if hasattr(chunk, "usage_metadata") and chunk.usage_metadata:
            self._update_statistics(
                input_tokens=chunk.usage_metadata.prompt_token_count,
                output_tokens=chunk.usage_metadata.candidates_token_count,
            )

        if (
            not chunk.candidates
            or not chunk.candidates[0].content
            or not chunk.candidates[0].content.parts
        ):
            return None

        if text := self._extract_text_from_parts(chunk.candidates[0].content.parts):
            return self.chunk_factory.text(text, delta=text)
        return None



# ruff: noqa: ARG002, SLF001, S105, S106, PERF401
"""Tests for OpenAIChatCompletionsProcessor.

Covers init, input validation, request building, message conversion,
stream/sync response handling, and completion chunk creation.
"""

from __future__ import annotations

import asyncio
from types import SimpleNamespace
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from openai import AsyncStream

from appkit_assistant.backend.processors.openai_chat_completion_processor import (
    OpenAIChatCompletionsProcessor,
)
from appkit_assistant.backend.schemas import (
    AIModel,
    ChunkType,
    Message,
    MessageType,
)

_PATCH = "appkit_assistant.backend.processors.openai_chat_completion_processor"


# ============================================================================
# Helpers
# ============================================================================


def _model(
    model_id: str = "gpt-4o",
    stream: bool = True,
    temperature: float = 0.7,
) -> AIModel:
    return AIModel(
        id=model_id,
        text=model_id,
        model=model_id,
        stream=stream,
        temperature=temperature,
    )


def _make_processor(
    api_key: str = "sk-test",
    base_url: str | None = None,
    on_azure: bool = False,
    models: dict[str, AIModel] | None = None,
) -> OpenAIChatCompletionsProcessor:
    if models is None:
        models = {"gpt-4o": _model()}
    with patch(
        "appkit_assistant.backend.processors.openai_base.AsyncOpenAI"
    ) as mock_cls:
        mock_cls.return_value = MagicMock()
        return OpenAIChatCompletionsProcessor(
            models=models,
            api_key=api_key,
            base_url=base_url,
            on_azure=on_azure,
        )


# ============================================================================
# Initialization
# ============================================================================


class TestInitialization:
    def test_basic_init(self) -> None:
        proc = _make_processor()
        assert proc.client is not None

    def test_no_api_key(self) -> None:
        proc = OpenAIChatCompletionsProcessor(models={"m": _model()}, api_key=None)
        assert proc.client is None

    def test_get_supported_models_with_key(self) -> None:
        proc = _make_processor()
        assert proc.get_supported_models() == proc.models

    def test_get_supported_models_no_key(self) -> None:
        proc = OpenAIChatCompletionsProcessor(models={"m": _model()}, api_key=None)
        assert proc.get_supported_models() == {}


# ============================================================================
# _validate_inputs
# ============================================================================


class TestValidateInputs:
    def test_no_client_raises(self) -> None:
        proc = OpenAIChatCompletionsProcessor(models={"m": _model()}, api_key=None)
        with pytest.raises(ValueError, match="not initialized"):
            proc._validate_inputs("m", None)

    def test_unknown_model_raises(self) -> None:
        proc = _make_processor()
        with pytest.raises(ValueError, match="not supported"):
            proc._validate_inputs("unknown-model", None)

    def test_valid_model_passes(self) -> None:
        proc = _make_processor()
        proc._validate_inputs("gpt-4o", None)  # no exception

    def test_mcp_servers_warns(self) -> None:
        proc = _make_processor()
        server = MagicMock()
        proc._validate_inputs("gpt-4o", [server])  # logs warning but no error


# ============================================================================
# _build_request_kwargs
# ============================================================================


class TestBuildRequestKwargs:
    def test_streaming_model(self) -> None:
        proc = _make_processor()
        msgs = [Message(type=MessageType.HUMAN, text="hello")]
        model = _model(stream=True)
        kwargs = proc._build_request_kwargs(msgs, model, None)
        assert kwargs["model"] == "gpt-4o"
        assert kwargs["stream"] is True
        assert kwargs["temperature"] == 0.7
        assert "stream_options" in kwargs
        assert kwargs["stream_options"]["include_usage"] is True

    def test_non_streaming_model(self) -> None:
        proc = _make_processor()
        msgs = [Message(type=MessageType.HUMAN, text="hello")]
        model = _model(stream=False)
        kwargs = proc._build_request_kwargs(msgs, model, None)
        assert kwargs["stream"] is False
        assert "stream_options" not in kwargs

    def test_payload_passthrough(self) -> None:
        proc = _make_processor()
        msgs = [Message(type=MessageType.HUMAN, text="hello")]
        model = _model()
        payload = {"top_p": 0.9}
        kwargs = proc._build_request_kwargs(msgs, model, payload)
        assert kwargs["extra_body"] == {"top_p": 0.9}


# ============================================================================
# _convert_messages_to_openai_format
# ============================================================================


class TestConvertMessages:
    def test_basic_conversion(self) -> None:
        proc = _make_processor()
        msgs = [
            Message(type=MessageType.HUMAN, text="Hello"),
            Message(type=MessageType.ASSISTANT, text="Hi"),
        ]
        result = proc._convert_messages_to_openai_format(msgs)
        assert len(result) == 2
        assert result[0]["role"] == "user"
        assert result[0]["content"] == "Hello"
        assert result[1]["role"] == "assistant"
        assert result[1]["content"] == "Hi"

    def test_system_message(self) -> None:
        proc = _make_processor()
        msgs = [
            Message(type=MessageType.SYSTEM, text="Be helpful"),
            Message(type=MessageType.HUMAN, text="Hello"),
        ]
        result = proc._convert_messages_to_openai_format(msgs)
        assert len(result) == 2
        assert result[0]["role"] == "system"
        assert result[1]["role"] == "user"

    def test_consecutive_user_merge(self) -> None:
        proc = _make_processor()
        msgs = [
            Message(type=MessageType.HUMAN, text="First"),
            Message(type=MessageType.HUMAN, text="Second"),
        ]
        result = proc._convert_messages_to_openai_format(msgs)
        assert len(result) == 1
        assert "First" in result[0]["content"]
        assert "Second" in result[0]["content"]

    def test_consecutive_system_not_merged(self) -> None:
        proc = _make_processor()
        msgs = [
            Message(type=MessageType.SYSTEM, text="S1"),
            Message(type=MessageType.SYSTEM, text="S2"),
        ]
        result = proc._convert_messages_to_openai_format(msgs)
        assert len(result) == 2

    def test_empty_messages(self) -> None:
        proc = _make_processor()
        result = proc._convert_messages_to_openai_format([])
        assert result == []

    def test_none_messages(self) -> None:
        proc = _make_processor()
        result = proc._convert_messages_to_openai_format(None)
        assert result == []


# ============================================================================
# _create_text_chunk
# ============================================================================


class TestCreateTextChunk:
    def test_basic(self) -> None:
        proc = _make_processor()
        chunk = proc._create_text_chunk("hello", "gpt-4o")
        assert chunk.type == ChunkType.TEXT
        assert chunk.text == "hello"
        assert chunk.chunk_metadata["source"] == "chat_completions"
        assert chunk.chunk_metadata["streaming"] == "False"
        assert chunk.chunk_metadata["model"] == "gpt-4o"

    def test_streaming(self) -> None:
        proc = _make_processor()
        chunk = proc._create_text_chunk(
            "hello", "gpt-4o", stream=True, message_id="msg-123"
        )
        assert chunk.chunk_metadata["streaming"] == "True"
        assert chunk.chunk_metadata["message_id"] == "msg-123"


# ============================================================================
# _create_completion_chunk
# ============================================================================


class TestCreateCompletionChunk:
    def test_creates_completion(self) -> None:
        proc = _make_processor()
        proc._reset_statistics("gpt-4o")
        chunk = proc._create_completion_chunk()
        assert chunk.type == ChunkType.COMPLETION
        assert chunk.chunk_metadata["status"] == "response_complete"
        assert chunk.statistics is not None


# ============================================================================
# Stream response handling
# ============================================================================


class TestHandleStreamResponse:
    @pytest.mark.asyncio
    async def test_text_events(self) -> None:
        proc = _make_processor()
        proc._reset_statistics("gpt-4o")

        event1 = SimpleNamespace(
            choices=[SimpleNamespace(delta=SimpleNamespace(content="Hello"))],
            id="msg-1",
            usage=None,
        )
        event2 = SimpleNamespace(
            choices=[SimpleNamespace(delta=SimpleNamespace(content=" world"))],
            id="msg-1",
            usage=None,
        )

        async def mock_stream():
            yield event1
            yield event2

        chunks = []
        async for chunk in proc._handle_stream_response(mock_stream(), "gpt-4o", None):
            chunks.append(chunk)

        assert len(chunks) == 2
        assert chunks[0].text == "Hello"
        assert chunks[1].text == " world"

    @pytest.mark.asyncio
    async def test_cancellation(self) -> None:
        proc = _make_processor()
        proc._reset_statistics("gpt-4o")

        event = SimpleNamespace(
            choices=[SimpleNamespace(delta=SimpleNamespace(content="X"))],
            id="msg-1",
            usage=None,
        )

        cancel = asyncio.Event()
        cancel.set()  # already cancelled

        async def mock_stream():
            yield event
            yield event

        chunks = []
        async for chunk in proc._handle_stream_response(
            mock_stream(), "gpt-4o", cancel
        ):
            chunks.append(chunk)

        assert len(chunks) == 0

    @pytest.mark.asyncio
    async def test_usage_captured(self) -> None:
        proc = _make_processor()
        proc._reset_statistics("gpt-4o")

        usage = SimpleNamespace(prompt_tokens=100, completion_tokens=50)
        event = SimpleNamespace(
            choices=[SimpleNamespace(delta=SimpleNamespace(content="T"))],
            id="msg-1",
            usage=usage,
        )

        async def mock_stream():
            yield event

        chunks = []
        async for chunk in proc._handle_stream_response(mock_stream(), "gpt-4o", None):
            chunks.append(chunk)

        stats = proc._get_statistics()
        assert stats.input_tokens == 100
        assert stats.output_tokens == 50


# ============================================================================
# Sync response handling
# ============================================================================


class TestHandleSyncResponse:
    @pytest.mark.asyncio
    async def test_basic_response(self) -> None:
        proc = _make_processor()
        proc._reset_statistics("gpt-4o")

        response = SimpleNamespace(
            choices=[SimpleNamespace(message=SimpleNamespace(content="Final answer"))],
            id="chatcmpl-123",
            usage=SimpleNamespace(prompt_tokens=10, completion_tokens=5),
        )

        chunks = []
        async for chunk in proc._handle_sync_response(response, "gpt-4o"):
            chunks.append(chunk)

        assert len(chunks) == 1
        assert chunks[0].text == "Final answer"
        stats = proc._get_statistics()
        assert stats.input_tokens == 10

    @pytest.mark.asyncio
    async def test_no_content(self) -> None:
        proc = _make_processor()
        proc._reset_statistics("gpt-4o")

        response = SimpleNamespace(
            choices=[SimpleNamespace(message=SimpleNamespace(content=None))],
            id="chatcmpl-123",
            usage=SimpleNamespace(prompt_tokens=5, completion_tokens=0),
        )

        chunks = []
        async for chunk in proc._handle_sync_response(response, "gpt-4o"):
            chunks.append(chunk)

        assert len(chunks) == 0

    @pytest.mark.asyncio
    async def test_no_usage(self) -> None:
        proc = _make_processor()
        proc._reset_statistics("gpt-4o")

        response = SimpleNamespace(
            choices=[SimpleNamespace(message=SimpleNamespace(content="Result"))],
            id="chatcmpl-123",
            usage=None,
        )

        chunks = []
        async for chunk in proc._handle_sync_response(response, "gpt-4o"):
            chunks.append(chunk)

        assert len(chunks) == 1


# ============================================================================
# Full process flow
# ============================================================================


class TestProcess:
    @pytest.mark.asyncio
    async def test_process_streaming(self) -> None:
        proc = _make_processor()
        proc._reset_statistics("gpt-4o")

        # Test streaming path by mocking the API and isinstance check
        msg_chunk = SimpleNamespace(
            choices=[SimpleNamespace(delta=SimpleNamespace(content="Hi"))],
            id="msg-1",
            usage=None,
        )
        completion_choice = SimpleNamespace(delta=SimpleNamespace(content=None))
        completion_chunk = SimpleNamespace(
            choices=[completion_choice],
            id="msg-1",
            usage=SimpleNamespace(prompt_tokens=10, completion_tokens=5),
        )

        # Create an async generator simulating OpenAI stream
        async def mock_stream_generator():
            yield msg_chunk
            yield completion_chunk

        # Mock the client.chat.completions.create to return the generator
        proc.client.chat.completions.create = AsyncMock(
            return_value=mock_stream_generator()
        )

        # Patch isinstance to recognize our generator as AsyncStream
        original_isinstance = isinstance

        def mocked_isinstance(obj, classinfo):
            # For AsyncStream checks on our mock generator, return True
            if classinfo is AsyncStream and hasattr(obj, "__anext__"):
                return True
            return original_isinstance(obj, classinfo)

        with patch(f"{_PATCH}.isinstance", side_effect=mocked_isinstance):
            msgs = [Message(type=MessageType.HUMAN, text="Hello")]
            chunks = []
            async for chunk in proc.process(msgs, "gpt-4o"):
                chunks.append(chunk)

            # Verify streaming path was taken and produced chunks
            assert any(c.type == ChunkType.TEXT for c in chunks), (
                "Should have TEXT chunks from streaming"
            )
            assert chunks[-1].type == ChunkType.COMPLETION, (
                "Last chunk should be COMPLETION"
            )



"""Tests for SystemPromptCache.

Covers caching, TTL-based invalidation, manual invalidation,
and error handling when no prompt exists.
"""

from datetime import UTC, datetime, timedelta
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from appkit_assistant.backend.system_prompt_cache import (
    CACHE_TTL_SECONDS,
    SystemPromptCache,
)


@pytest.fixture
def fresh_cache() -> SystemPromptCache:
    """Create a fresh cache instance by resetting the singleton."""
    SystemPromptCache._instance = None
    return SystemPromptCache()


@pytest.fixture(autouse=True)
def _reset_singleton():
    """Reset singleton after each test."""
    yield
    SystemPromptCache._instance = None


class TestCacheInit:
    def test_singleton(self) -> None:
        SystemPromptCache._instance = None
        c1 = SystemPromptCache()
        c2 = SystemPromptCache()
        assert c1 is c2

    def test_defaults(self, fresh_cache: SystemPromptCache) -> None:
        assert fresh_cache.is_cached is False
        assert fresh_cache.cached_version is None
        assert fresh_cache._ttl_seconds == CACHE_TTL_SECONDS


class TestCacheValidity:
    def test_empty_cache_invalid(self, fresh_cache: SystemPromptCache) -> None:
        assert fresh_cache._is_cache_valid() is False

    def test_valid_cache(self, fresh_cache: SystemPromptCache) -> None:
        fresh_cache._cached_prompt = "test"
        fresh_cache._cached_version = 1
        fresh_cache._cache_timestamp = datetime.now(UTC)
        assert fresh_cache._is_cache_valid() is True

    def test_expired_cache(self, fresh_cache: SystemPromptCache) -> None:
        fresh_cache._cached_prompt = "test"
        fresh_cache._cached_version = 1
        fresh_cache._cache_timestamp = datetime.now(UTC) - timedelta(
            seconds=CACHE_TTL_SECONDS + 10
        )
        assert fresh_cache._is_cache_valid() is False


class TestGetPrompt:
    @pytest.mark.asyncio
    async def test_cache_hit(self, fresh_cache: SystemPromptCache) -> None:
        fresh_cache._cached_prompt = "cached prompt"
        fresh_cache._cached_version = 1
        fresh_cache._cache_timestamp = datetime.now(UTC)

        result = await fresh_cache.get_prompt()
        assert result == "cached prompt"

    @pytest.mark.asyncio
    async def test_cache_miss_loads_from_db(
        self, fresh_cache: SystemPromptCache
    ) -> None:
        mock_prompt = MagicMock()
        mock_prompt.prompt = "db prompt"
        mock_prompt.version = 2

        mock_session = AsyncMock()
        mock_repo = MagicMock()
        mock_repo.find_latest = AsyncMock(return_value=mock_prompt)

        with (
            patch(
                "appkit_assistant.backend.system_prompt_cache.get_asyncdb_session"
            ) as mock_ctx,
            patch(
                "appkit_assistant.backend.system_prompt_cache.system_prompt_repo",
                mock_repo,
            ),
        ):
            mock_ctx.return_value.__aenter__ = AsyncMock(return_value=mock_session)
            mock_ctx.return_value.__aexit__ = AsyncMock(return_value=False)

            result = await fresh_cache.get_prompt()
            assert result == "db prompt"
            assert fresh_cache._cached_version == 2
            assert fresh_cache._cached_prompt == "db prompt"

    @pytest.mark.asyncio
    async def test_no_prompt_raises(self, fresh_cache: SystemPromptCache) -> None:
        mock_repo = MagicMock()
        mock_repo.find_latest = AsyncMock(return_value=None)

        with (
            patch(
                "appkit_assistant.backend.system_prompt_cache.get_asyncdb_session"
            ) as mock_ctx,
            patch(
                "appkit_assistant.backend.system_prompt_cache.system_prompt_repo",
                mock_repo,
            ),
        ):
            mock_ctx.return_value.__aenter__ = AsyncMock(return_value=AsyncMock())
            mock_ctx.return_value.__aexit__ = AsyncMock(return_value=False)

            with pytest.raises(ValueError, match="No system prompt found"):
                await fresh_cache.get_prompt()


class TestInvalidate:
    @pytest.mark.asyncio
    async def test_invalidate_clears_cache(
        self, fresh_cache: SystemPromptCache
    ) -> None:
        fresh_cache._cached_prompt = "test"
        fresh_cache._cached_version = 1
        fresh_cache._cache_timestamp = datetime.now(UTC)

        await fresh_cache.invalidate()
        assert fresh_cache._cached_prompt is None
        assert fresh_cache._cached_version is None
        assert fresh_cache._cache_timestamp is None

    @pytest.mark.asyncio
    async def test_invalidate_empty_cache_no_error(
        self, fresh_cache: SystemPromptCache
    ) -> None:
        await fresh_cache.invalidate()  # Should not raise


class TestSetTtl:
    def test_set_ttl(self, fresh_cache: SystemPromptCache) -> None:
        fresh_cache.set_ttl(60)
        assert fresh_cache._ttl_seconds == 60


class TestProperties:
    def test_is_cached_false_when_empty(self, fresh_cache: SystemPromptCache) -> None:
        assert fresh_cache.is_cached is False

    def test_is_cached_true_when_valid(self, fresh_cache: SystemPromptCache) -> None:
        fresh_cache._cached_prompt = "test"
        fresh_cache._cached_version = 1
        fresh_cache._cache_timestamp = datetime.now(UTC)
        assert fresh_cache.is_cached is True

    def test_cached_version_when_valid(self, fresh_cache: SystemPromptCache) -> None:
        fresh_cache._cached_prompt = "test"
        fresh_cache._cached_version = 5
        fresh_cache._cache_timestamp = datetime.now(UTC)
        assert fresh_cache.cached_version == 5

    def test_cached_version_none_when_expired(
        self, fresh_cache: SystemPromptCache
    ) -> None:
        fresh_cache._cached_prompt = "test"
        fresh_cache._cached_version = 5
        fresh_cache._cache_timestamp = datetime.now(UTC) - timedelta(
            seconds=CACHE_TTL_SECONDS + 10
        )
        assert fresh_cache.cached_version is None



"""MCP Apps Service for direct MCP client communication.

Provides a parallel direct MCP client connection to MCP servers,
alongside the existing LLM-provider-mediated path. The direct client
negotiates the io.modelcontextprotocol/ui extension (SEP-1865 §Capability
Negotiation), discovers UI-enabled tools, fetches ui:// HTML resources,
and proxies tool calls from app iframes.
"""

import json
import logging
import time
from contextlib import asynccontextmanager
from typing import Any

import httpx
from mcp import ClientSession  # noqa: F401 - kept for re-export / type hints
from mcp import types as t
from mcp.client.streamable_http import streamable_http_client
from mcp.types import (
    CallToolResult,
    ClientCapabilities,
    Implementation,
    InitializeResult,
    Tool,
)

from appkit_assistant.backend.database.models import MCPServer
from appkit_assistant.backend.schemas import (
    McpAppResource,
    McpAppToolInfo,
    MCPAuthType,
)
from appkit_assistant.backend.services.mcp_token_service import (
    MCPTokenService,
)

logger = logging.getLogger(__name__)

# Default TTL for cached tool metadata (seconds)
_TOOL_CACHE_TTL_S = 300

# Extension identifier per SEP-1865
_EXTENSION_ID = "io.modelcontextprotocol/ui"
_SUPPORTED_MIME_TYPE = "text/html;profile=mcp-app"

# Client info sent to MCP servers
_CLIENT_INFO = Implementation(name="appkit", version="1.0.0")


class _McpAppsClientSession(ClientSession):
    """ClientSession with MCP Apps capability advertised.

    Overrides `initialize()` to include the ``io.modelcontextprotocol/ui``
    extension in the ``experimental`` field of ``ClientCapabilities``.
    This signals to the MCP server that this host supports MCP Apps,
    so the server can register UI-enabled tools (spec §Capability Negotiation).
    """

    async def initialize(self) -> InitializeResult:
        result = await self.send_request(
            t.ClientRequest(
                t.InitializeRequest(
                    params=t.InitializeRequestParams(
                        protocolVersion=t.LATEST_PROTOCOL_VERSION,
                        capabilities=ClientCapabilities(
                            experimental={
                                "extensions": {
                                    _EXTENSION_ID: {
                                        "mimeTypes": [_SUPPORTED_MIME_TYPE],
                                    }
                                }
                            },
                        ),
                        clientInfo=_CLIENT_INFO,
                    ),
                )
            ),
            t.InitializeResult,
        )
        self._server_capabilities = result.capabilities  # type: ignore[attr-defined]
        await self.send_notification(t.ClientNotification(t.InitializedNotification()))
        logger.debug(
            "MCP Apps session initialized (protocolVersion=%s)",
            result.protocolVersion,
        )
        return result


class McpAppsService:
    """Service for direct MCP client communication with App support.

    Manages MCP client sessions, discovers UI-enabled tools,
    fetches resources, and proxies tool calls for MCP Apps.

    The tool cache is a class-level dict shared across all instances so
    that discovery results survive across message submissions.
    """

    # Class-level cache: (server_id, user_id) -> (tools, timestamp)
    _tool_cache: dict[tuple[int, int], tuple[list[McpAppToolInfo], float]] = {}

    def __init__(self, token_service: MCPTokenService | None = None) -> None:
        self._token_service = token_service

    @asynccontextmanager
    async def _connect_for_apps(self, server: MCPServer, user_id: int):
        """Create an authenticated MCP Apps client session context.

        Args:
            server: The MCP server configuration
            user_id: The user's ID

        Yields:
            An initialized _McpAppsClientSession
        """
        headers = await self._get_auth_headers(server, user_id)
        async with (
            httpx.AsyncClient(headers=headers) as http_client,
            streamable_http_client(server.url, http_client=http_client) as (
                read_stream,
                write_stream,
                _,
            ),
            # Use _McpAppsClientSession to advertise io.modelcontextprotocol/ui
            _McpAppsClientSession(read_stream, write_stream) as session,
        ):
            await session.initialize()
            yield session

    def _get_cached_tools(
        self,
        key: tuple[int, int],
    ) -> list[McpAppToolInfo] | None:
        """Retrieve cached UI tools if valid."""
        if cached := self._tool_cache.get(key):
            tools, ts = cached
            if (time.monotonic() - ts) < _TOOL_CACHE_TTL_S:
                return tools
        return None

    async def _get_auth_headers(
        self,
        server: MCPServer,
        user_id: int,
    ) -> dict[str, str]:
        """Build authorization headers for a server request."""
        headers: dict[str, str] = {}

        # Parse custom headers from server configuration (API_KEY etc.)
        if server.headers:
            try:
                headers_dict = json.loads(server.headers)
                if isinstance(headers_dict, dict):
                    headers.update(headers_dict)
            except json.JSONDecodeError:
                logger.warning("Invalid headers JSON for server %s", server.name)

        # Forward the user identity so MCP servers can scope data
        if user_id > 0:
            headers["x-user-id"] = str(user_id)

        # Override with OAuth token if available
        if server.auth_type == MCPAuthType.OAUTH_DISCOVERY and self._token_service:
            token = await self._token_service.get_valid_token(server, user_id)
            if token:
                headers["Authorization"] = f"Bearer {token.access_token}"
                logger.debug(
                    "Injected OAuth token for MCP Apps server %s",
                    server.name,
                )

        return headers

    async def discover_ui_tools(
        self,
        server: MCPServer,
        user_id: int,
    ) -> list[McpAppToolInfo]:
        """Discover tools with UI (App) views on an MCP server."""
        if server.id is None:
            return []

        cache_key = (server.id, user_id)
        if tools := self._get_cached_tools(cache_key):
            return tools

        try:
            async with self._connect_for_apps(server, user_id) as session:
                ui_tools = await self._list_ui_tools(session, server)

            self._tool_cache[cache_key] = (ui_tools, time.monotonic())
            logger.info(
                "Discovered %d UI tools on server %s",
                len(ui_tools),
                server.name,
            )
            return ui_tools
        except Exception:
            logger.exception(
                "Failed to discover UI tools on server %s",
                server.name,
            )
            return []

    async def _list_ui_tools(
        self,
        session: ClientSession,
        server: MCPServer,
    ) -> list[McpAppToolInfo]:
        """Fetch and filter UI-enabled tools from an active session."""
        result = await session.list_tools()

        return [
            info
            for tool in result.tools
            if (info := self._extract_ui_tool_info(tool, server))
        ]

    def _extract_ui_tool_info(
        self,
        tool: Tool,
        server: MCPServer,
    ) -> McpAppToolInfo | None:
        """Extract UI tool info from an MCP Tool definition."""
        meta = getattr(tool, "meta", None) or {}
        ui_meta: dict[str, Any] = meta.get("ui", {}) if meta else {}
        resource_uri = ui_meta.get("resourceUri")

        if not resource_uri:
            return None

        return McpAppToolInfo(
            tool_name=tool.name,
            resource_uri=resource_uri,
            visibility=ui_meta.get("visibility", []),
            server_id=server.id or 0,
            server_label=server.name,
            input_schema=(tool.inputSchema or {}),
        )

    async def fetch_resource(
        self,
        server: MCPServer,
        user_id: int,
        resource_uri: str,
    ) -> McpAppResource | None:
        """Fetch an MCP App resource (HTML content) from a server."""
        try:
            async with self._connect_for_apps(server, user_id) as session:
                result = await session.read_resource(resource_uri)

            html_content = ""
            csp: dict[str, Any] | None = None
            permissions: dict[str, bool] | None = None
            prefers_border: bool | None = None

            for content in result.contents:
                if hasattr(content, "text"):
                    html_content += content.text

                # Extract _meta.ui fields from resource content
                raw_meta: dict[str, Any] = getattr(content, "_meta", None) or {}
                if ui_meta := raw_meta.get("ui", {}):
                    csp = ui_meta.get("csp", csp)
                    permissions = ui_meta.get("permissions", permissions)
                    if "prefersBorder" in ui_meta:
                        prefers_border = bool(ui_meta["prefersBorder"])

            return McpAppResource(
                uri=resource_uri,
                html_content=html_content,
                csp=csp,
                permissions=permissions,
                prefers_border=prefers_border,
            )
        except Exception:
            logger.exception(
                "Failed to fetch resource %s from server %s",
                resource_uri,
                server.name,
            )
            return None

    async def proxy_tool_call(
        self,
        server: MCPServer,
        user_id: int,
        tool_name: str,
        arguments: dict[str, Any],
    ) -> dict[str, Any]:
        """Proxy a tool call from an MCP App iframe to the server."""
        logger.info(
            "Proxying tool call %s to server %s for user %d",
            tool_name,
            server.name,
            user_id,
        )

        try:
            async with self._connect_for_apps(server, user_id) as session:
                result: CallToolResult = await session.call_tool(tool_name, arguments)

            # Note: CallToolResult needs standardizing to dict
            # We use the existing function outside the class or standard extraction
            return _call_tool_result_to_dict(result)
        except Exception:
            logger.exception(
                "Failed to proxy tool call %s to server %s",
                tool_name,
                server.name,
            )
            return {"isError": True, "content": []}

    async def is_apps_supported(
        self,
        server: MCPServer,
        user_id: int,
    ) -> bool:
        """Check if the server supports MCP Apps (has UI-enabled tools)."""
        if server.id is None:
            return False

        # Reuse discovery logic which handles caching.
        # We removed _apps_support_cache as it was redundant with _tool_cache
        tools = await self.discover_ui_tools(server, user_id)
        return len(tools) > 0

    def get_cached_ui_tools(
        self,
        server_id: int,
        user_id: int,
    ) -> list[McpAppToolInfo]:
        """Get cached UI tools without making a network request."""
        return self._get_cached_tools((server_id, user_id)) or []

    def build_ui_tool_registry(
        self,
        tools: list[McpAppToolInfo],
    ) -> dict[str, McpAppToolInfo]:
        """Build a tool name → McpAppToolInfo mapping."""
        return {tool.tool_name: tool for tool in tools}


def _call_tool_result_to_dict(result: CallToolResult) -> dict[str, Any]:
    """Convert a CallToolResult to a serializable dictionary."""
    content_list = [item.model_dump(exclude_none=True) for item in result.content]

    return {
        "isError": bool(result.isError),
        "content": content_list,
    }



"""State management for user skill selections."""

import logging
from collections.abc import AsyncGenerator
from typing import Any

import reflex as rx

from appkit_assistant.backend.database.repositories import (
    skill_repo,
    user_skill_repo,
)
from appkit_commons.database.session import get_asyncdb_session
from appkit_user.authentication.states import UserSession

logger = logging.getLogger(__name__)


class SkillState(rx.State):
    """State for user-scoped skill enable/disable preferences."""

    available_skills: list[dict[str, Any]] = []
    loading: bool = False

    async def load_user_skills(self) -> AsyncGenerator[Any, Any]:
        """Load active skills with the user's enable/disable state."""
        self.loading = True
        yield
        try:
            session_state = await self.get_state(UserSession)
            user_id = session_state.user_id
            if not user_id:
                self.available_skills = []
                return

            async with get_asyncdb_session() as session:
                active_skills = await skill_repo.find_all_active_ordered_by_name(
                    session
                )
                selections = await user_skill_repo.find_by_user_id(session, user_id)

            enabled_map: dict[str, bool] = {
                sel.skill_openai_id: sel.enabled for sel in selections
            }

            self.available_skills = [
                {
                    "openai_id": s.openai_id,
                    "name": s.name,
                    "description": s.description or "",
                    "enabled": enabled_map.get(s.openai_id, False),
                }
                for s in active_skills
            ]
            logger.debug(
                "Loaded %d skills for user %d",
                len(self.available_skills),
                user_id,
            )
        except Exception as e:
            logger.error("Failed to load user skills: %s", e)
            yield rx.toast.error(
                "Fehler beim Laden der Skills.",
                position="top-right",
            )
        finally:
            self.loading = False

    async def toggle_skill_enabled(
        self, openai_id: str, enabled: bool
    ) -> AsyncGenerator[Any, Any]:
        """Toggle the enabled state of a skill for the current user."""
        # Optimistic update
        for i, skill in enumerate(self.available_skills):
            if skill["openai_id"] == openai_id:
                self.available_skills[i]["enabled"] = enabled
                break
        yield

        try:
            session_state = await self.get_state(UserSession)
            user_id = session_state.user_id
            if not user_id:
                return

            async with get_asyncdb_session() as session:
                await user_skill_repo.upsert(session, user_id, openai_id, enabled)

            status = "aktiviert" if enabled else "deaktiviert"
            yield rx.toast.info(
                f"Skill {status}.",
                position="top-right",
            )
        except Exception as e:
            logger.error("Failed to toggle skill %s: %s", openai_id, e)
            # Revert optimistic update
            for i, skill in enumerate(self.available_skills):
                if skill["openai_id"] == openai_id:
                    self.available_skills[i]["enabled"] = not enabled
                    break
            yield rx.toast.error(
                "Fehler beim Ändern des Skill-Status.",
                position="top-right",
            )

    @rx.var
    def has_skills(self) -> bool:
        """Whether any skills are available."""
        return len(self.available_skills) > 0

    @rx.var
    def enabled_skill_ids(self) -> list[str]:
        """Return a list of enabled skill OpenAI IDs."""
        return [s["openai_id"] for s in self.available_skills if s.get("enabled")]



"""MCP server tool selection mixin for ThreadState.

Handles loading, toggling, and applying MCP server selections.
"""

import logging

import reflex as rx

from appkit_assistant.backend.database.models import MCPServer
from appkit_assistant.backend.database.repositories import mcp_server_repo
from appkit_commons.database.session import get_asyncdb_session
from appkit_user.authentication.states import UserSession

logger = logging.getLogger(__name__)


class McpToolsMixin:
    """Mixin for MCP server (tool) selection.

    Expects state vars: ``selected_mcp_servers``, ``show_tools_modal``,
    ``available_mcp_servers``, ``temp_selected_mcp_servers``,
    ``server_selection_state``.
    """

    @rx.event
    async def load_mcp_servers(self) -> None:
        """Load available active MCP servers filtered by user roles."""
        user_session = await self.get_state(UserSession)
        user = await user_session.authenticated_user
        user_roles: list[str] = user.roles if user else []

        async with get_asyncdb_session() as session:
            servers = await mcp_server_repo.find_all_active_ordered_by_name(session)
            filtered_servers = [
                MCPServer(**s.model_dump())
                for s in servers
                if not s.required_role or s.required_role in user_roles
            ]
            self.available_mcp_servers = filtered_servers

    @rx.event
    def toggle_tools_modal(self, show: bool) -> None:
        """Set the visibility of the tools modal."""
        self.show_tools_modal = show

    @rx.event
    def toggle_mcp_server_selection(self, server_id: int, selected: bool) -> None:
        """Toggle MCP server selection in the modal."""
        self.server_selection_state[server_id] = selected
        if selected and server_id not in self.temp_selected_mcp_servers:
            self.temp_selected_mcp_servers.append(server_id)
        elif not selected and server_id in self.temp_selected_mcp_servers:
            self.temp_selected_mcp_servers.remove(server_id)

    @rx.event
    def apply_mcp_server_selection(self) -> None:
        """Apply the temporary MCP server selection."""
        self.selected_mcp_servers = [
            server
            for server in self.available_mcp_servers
            if server.id in self.temp_selected_mcp_servers
        ]
        self.show_tools_modal = False

    @rx.event
    def deselect_all_mcp_servers(self) -> None:
        """Deselect all MCP servers in the modal."""
        self.server_selection_state = {}
        self.temp_selected_mcp_servers = []

    @rx.event
    def is_mcp_server_selected(self, server_id: int) -> bool:
        """Check if an MCP server is selected."""
        return server_id in self.temp_selected_mcp_servers

    def _restore_mcp_selection(self, server_ids: list[int]) -> None:
        """Restore MCP selection state from a list of server IDs."""
        if not server_ids:
            self.selected_mcp_servers = []
            self.temp_selected_mcp_servers = []
            self.server_selection_state = {}
            return

        self.selected_mcp_servers = [
            server for server in self.available_mcp_servers if server.id in server_ids
        ]
        self.temp_selected_mcp_servers = list(server_ids)
        self.server_selection_state = dict.fromkeys(server_ids, True)



"""Component for slash command palette in the composer."""

import reflex as rx

import appkit_mantine as mn
from appkit_assistant.backend.schemas import CommandDefinition
from appkit_assistant.state.thread_state import ThreadState
from appkit_assistant.state.user_prompt_state import UserPromptState

_GRID_TEMPLATE = "max-content minmax(0, 1fr) auto"
_GRID_PROPS = {
    "display": "grid",
    "grid_template_columns": _GRID_TEMPLATE,
    "gap": "2px",
    "width": "100%",
}


def render_command_label(command: CommandDefinition, index: int) -> rx.Component:
    """Render the label cell for a command item."""
    is_selected = ThreadState.selected_command_index == index
    return mn.text(
        command.label,
        fw=600,
        size="sm",
        color=rx.cond(is_selected, rx.color("accent", 11), "inherit"),
        truncate="end",
    )


def render_command_description(command: CommandDefinition) -> rx.Component:
    """Render the description cell for a command item."""
    return mn.text(
        command.description,
        size="xs",
        c="dimmed",
        truncate="end",
    )


def render_edit_button(command: CommandDefinition) -> rx.Component:
    """Render the edit button for editable commands."""
    return rx.cond(
        command.is_editable,
        mn.action_icon(
            rx.icon("pencil", size=14),
            on_click=[
                rx.stop_propagation,
                lambda: UserPromptState.open_edit_modal(command.id),
            ],
            variant="subtle",
            c="gray",
            size="sm",
            radius="sm",
            p="4px",
        ),
        mn.box(w="24px"),
    )


def render_command_item(command: CommandDefinition, index: int) -> rx.Component:
    """Render a single command item in the palette."""
    is_selected = ThreadState.selected_command_index == index

    return mn.box(
        render_command_label(command, index),
        render_command_description(command),
        render_edit_button(command),
        data_selected=rx.cond(is_selected, "true", "false"),
        class_name="command-palette-item",
        display="grid",
        grid_template_columns="subgrid",
        grid_column="1 / -1",
        p="8px 12px",
        border_radius="var(--mantine-radius-sm)",
        cursor="pointer",
        gap="12px",
        align_items="center",
        mw="0",
        w="100%",
        background=rx.cond(
            is_selected,
            rx.color("accent", 3),
            "transparent",
        ),
        _hover={"background": rx.color("gray", 3)},
        on_click=lambda: ThreadState.select_command(command.id),
    )


def render_new_prompt_item() -> rx.Component:
    """Render the 'New Prompt...' entry at the bottom of the palette."""
    return mn.box(
        mn.button(
            mn.text(
                "Neuer Prompt...",
                size="xs",
                fw=600,
                c="dimmed",
                truncate="end",
            ),
            left_section=rx.icon("plus", size=14, color=rx.color("gray", 10)),
            gap="4px",
            align="center",
            variant="subtle",
            size="xs",
            on_click=UserPromptState.open_new_modal,
        ),
        rx.spacer(),
        mn.action_icon(
            rx.icon("x", size=16),
            on_click=[rx.stop_propagation, ThreadState.dismiss_command_palette],
            variant="subtle",
            c="gray",
            size="xs",
            radius="xs",
            p="0",
        ),
        display="flex",
        align_items="center",
        gap="12px",
        p="8px 12px 8px 3px",
        br="6px",
        w="100%",
        mb="2px",
    )


def render_section_header(label: str) -> rx.Component:
    """Render a section header for command groups."""
    return mn.group(
        mn.divider(style={"flex": 1}, m=0),
        mn.text(
            label,
            size="xs",
            color=rx.color("gray", 9),
            truncate="end",
            fw=500,
        ),
        mn.divider(style={"flex": 1}, m=0),
        align="center",
        w="100%",
        # gap="xs",
    )


def _render_command_grid(
    commands: list[CommandDefinition], offset: int = 0
) -> rx.Component:
    """Render a list of commands in a grid layout."""
    return mn.box(
        rx.foreach(
            commands,
            lambda cmd, idx: render_command_item(cmd, idx + offset),
        ),
        **_GRID_PROPS,
    )


def command_palette_content() -> rx.Component:
    """Render the command palette content."""
    return mn.stack(
        rx.cond(
            ThreadState.filtered_commands.length() > 0,
            rx.fragment(
                rx.cond(
                    ThreadState.has_filtered_user_prompts,
                    _render_command_grid(ThreadState.filtered_user_prompts),
                    rx.fragment(),
                ),
                rx.cond(
                    ThreadState.has_filtered_shared_prompts,
                    rx.fragment(
                        render_section_header("Geteilte Prompts"),
                        _render_command_grid(
                            ThreadState.filtered_shared_prompts,
                            offset=ThreadState.filtered_user_prompts.length(),
                        ),
                    ),
                    rx.fragment(),
                ),
            ),
            mn.box(
                mn.text(
                    "Keine Befehle gefunden",
                    size="sm",
                    c="dimmed",
                ),
                p="12px",
                text_align="center",
            ),
        ),
        gap=0,
        w="100%",
        p="4px",
    )


def command_palette() -> rx.Component:
    """Render the command palette popover."""
    return rx.cond(
        ThreadState.show_command_palette,
        rx.fragment(
            # Overlay to catch clicks outside
            mn.box(
                pos="fixed",
                inset="0",
                z_index="999",
                on_click=ThreadState.dismiss_command_palette,
            ),
            mn.box(
                mn.card(
                    mn.scroll_area.autosize(
                        render_new_prompt_item(),
                        command_palette_content(),
                        type="auto",
                        scrollbars="y",
                        max_height="276px",
                    ),
                    shadow="md",
                    padding="xs",
                    radius="md",
                    with_border=True,
                ),
                id="command-palette",
                pos="absolute",
                bottom="100%",
                left="9px",
                miw="420px",
                w="max-content",
                maw="50%",
                mb="-6px",
                style={"zIndex": "1000"},
            ),
        ),
        rx.fragment(),
    )



"""Tests for FileValidationService."""

import tempfile
from pathlib import Path

from appkit_assistant.backend.services.file_validation import (
    FileValidationService,
    get_file_validation_service,
)


class TestFileValidationService:
    """Test suite for FileValidationService."""

    def test_get_file_extension_extracts_lowercase(self) -> None:
        """get_file_extension returns lowercase extension."""
        service = FileValidationService()

        assert service.get_file_extension("file.PDF") == "pdf"
        assert service.get_file_extension("document.DOCX") == "docx"

    def test_get_file_extension_handles_multiple_dots(self) -> None:
        """get_file_extension returns last extension only."""
        service = FileValidationService()

        assert service.get_file_extension("archive.tar.gz") == "gz"
        assert service.get_file_extension("file.backup.pdf") == "pdf"

    def test_get_file_extension_returns_empty_when_no_extension(self) -> None:
        """get_file_extension returns empty string for no extension."""
        service = FileValidationService()

        assert service.get_file_extension("README") == ""
        assert service.get_file_extension("file_without_ext") == ""

    def test_is_image_file_detects_image_extensions(self) -> None:
        """is_image_file returns True for image files."""
        service = FileValidationService()

        assert service.is_image_file("photo.png") is True
        assert service.is_image_file("image.JPG") is True
        assert service.is_image_file("picture.jpeg") is True
        assert service.is_image_file("graphic.gif") is True
        assert service.is_image_file("modern.webp") is True

    def test_is_image_file_rejects_non_images(self) -> None:
        """is_image_file returns False for non-image files."""
        service = FileValidationService()

        assert service.is_image_file("document.pdf") is False
        assert service.is_image_file("data.csv") is False
        assert service.is_image_file("text.md") is False

    def test_get_media_type_returns_correct_mime(self) -> None:
        """get_media_type returns correct MIME type for known extensions."""
        service = FileValidationService()

        assert service.get_media_type("file.pdf") == "application/pdf"
        assert service.get_media_type("image.png") == "image/png"
        assert service.get_media_type("photo.jpg") == "image/jpeg"
        assert service.get_media_type("data.csv") == "text/csv"
        assert (
            service.get_media_type("doc.docx")
            == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
        )

    def test_get_media_type_defaults_for_unknown(self) -> None:
        """get_media_type returns default for unknown extensions."""
        service = FileValidationService()

        assert service.get_media_type("file.xyz") == "application/octet-stream"
        assert service.get_media_type("unknown.bin") == "application/octet-stream"

    def test_validate_file_fails_when_not_exists(self) -> None:
        """validate_file returns False when file doesn't exist."""
        service = FileValidationService()

        valid, error = service.validate_file("/nonexistent/file.pdf")

        assert valid is False
        assert "not found" in error.lower()

    def test_validate_file_fails_for_unsupported_extension(self) -> None:
        """validate_file returns False for unsupported file types."""
        service = FileValidationService()

        with tempfile.NamedTemporaryFile(suffix=".exe", delete=False) as f:
            f.write(b"test data")
            temp_path = f.name

        try:
            valid, error = service.validate_file(temp_path)

            assert valid is False
            assert "unsupported" in error.lower()
            assert "exe" in error.lower()
        finally:
            Path(temp_path).unlink(missing_ok=True)

    def test_validate_file_fails_when_too_large(self) -> None:
        """validate_file returns False when file exceeds size limit."""
        service = FileValidationService()

        with tempfile.NamedTemporaryFile(suffix=".pdf", delete=False) as f:
            # Write > 5MB
            f.write(b"x" * (6 * 1024 * 1024))
            temp_path = f.name

        try:
            valid, error = service.validate_file(temp_path)

            assert valid is False
            assert "too large" in error.lower()
            assert "max 5mb" in error.lower()
        finally:
            Path(temp_path).unlink(missing_ok=True)

    def test_validate_file_succeeds_for_valid_file(self) -> None:
        """validate_file returns True for valid file."""
        service = FileValidationService()

        with tempfile.NamedTemporaryFile(suffix=".pdf", delete=False) as f:
            f.write(b"test pdf content")
            temp_path = f.name

        try:
            valid, error = service.validate_file(temp_path)

            assert valid is True
            assert error == ""
        finally:
            Path(temp_path).unlink(missing_ok=True)

    def test_validate_file_allows_all_supported_extensions(self) -> None:
        """validate_file accepts all ALLOWED_EXTENSIONS."""
        service = FileValidationService()

        for ext in service.ALLOWED_EXTENSIONS:
            with tempfile.NamedTemporaryFile(suffix=f".{ext}", delete=False) as f:
                f.write(b"test data")
                temp_path = f.name

            try:
                valid, _ = service.validate_file(temp_path)
                assert valid is True, f"Extension {ext} should be allowed"
            finally:
                Path(temp_path).unlink(missing_ok=True)

    def test_validate_file_at_size_limit(self) -> None:
        """validate_file accepts files at exactly 5MB."""
        service = FileValidationService()

        with tempfile.NamedTemporaryFile(suffix=".pdf", delete=False) as f:
            # Write exactly 5MB
            f.write(b"x" * (5 * 1024 * 1024))
            temp_path = f.name

        try:
            valid, error = service.validate_file(temp_path)

            assert valid is True
            assert error == ""
        finally:
            Path(temp_path).unlink(missing_ok=True)


class TestFileValidationServiceSingleton:
    """Test suite for singleton pattern."""

    def test_get_file_validation_service_returns_instance(self) -> None:
        """get_file_validation_service returns FileValidationService."""
        service = get_file_validation_service()

        assert isinstance(service, FileValidationService)

    def test_get_file_validation_service_returns_same_instance(self) -> None:
        """get_file_validation_service returns singleton."""
        service1 = get_file_validation_service()
        service2 = get_file_validation_service()

        assert service1 is service2



import asyncio
import logging
from datetime import UTC, datetime, timedelta
from typing import Final, Self

from appkit_assistant.backend.database.repositories import system_prompt_repo
from appkit_commons.database.session import get_asyncdb_session

logger = logging.getLogger(__name__)

# Cache TTL in seconds (default: 5 minutes)
CACHE_TTL_SECONDS: Final[int] = 300


class SystemPromptCache:
    """Singleton cache for system prompt with TTL-based invalidation.

    Features:
    - Lazy loading on first access
    - Automatic cache invalidation after TTL expires
    - Thread-safe with asyncio lock
    - Manual invalidation support for immediate updates
    """

    _instance: "SystemPromptCache | None" = None
    _lock: asyncio.Lock = asyncio.Lock()

    def __new__(cls) -> Self:
        """Ensure singleton pattern."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False  # noqa: SLF001
        return cls._instance

    def __init__(self) -> None:
        """Initialize cache state (only once due to singleton)."""
        if self._initialized:
            return

        self._cached_prompt: str | None = None
        self._cached_version: int | None = None
        self._cache_timestamp: datetime | None = None
        self._ttl_seconds: int = CACHE_TTL_SECONDS
        self._initialized = True

        logger.debug(
            "SystemPromptCache initialized with TTL=%d seconds",
            self._ttl_seconds,
        )

    def _is_cache_valid(self) -> bool:
        """Check if cached prompt is still valid based on TTL."""
        if self._cached_prompt is None or self._cache_timestamp is None:
            return False

        elapsed = datetime.now(UTC) - self._cache_timestamp
        is_valid = elapsed < timedelta(seconds=self._ttl_seconds)

        if not is_valid:
            logger.debug("Cache expired after %s seconds", elapsed.total_seconds())

        return is_valid

    async def get_prompt(self) -> str:
        """Get the latest system prompt (from cache or database).

        Returns:
            The current system prompt text.

        Raises:
            ValueError: If no system prompt exists in database.
        """
        async with self._lock:
            if self._is_cache_valid():
                logger.debug(
                    "Cache hit: version=%d, age=%s seconds",
                    self._cached_version,
                    (datetime.now(UTC) - self._cache_timestamp).total_seconds(),
                )
                return self._cached_prompt

            # Cache miss or expired - fetch from database
            logger.debug("Cache miss - fetching latest prompt from database")

            async with get_asyncdb_session() as session:
                latest_prompt = await system_prompt_repo.find_latest(session)

                if latest_prompt is None:
                    # Raise inside to exit or handle outside
                    pass
                else:
                    # Capture values while attached
                    prompt_text = latest_prompt.prompt
                    prompt_version = latest_prompt.version

            if latest_prompt is None:
                msg = "No system prompt found in database"
                logger.error(msg)
                raise ValueError(msg)

            self._cached_prompt = prompt_text
            self._cached_version = prompt_version
            self._cache_timestamp = datetime.now(UTC)

            logger.debug(
                "Cached prompt version %d (%d characters)",
                self._cached_version,
                len(self._cached_prompt),
            )

            return self._cached_prompt

    async def invalidate(self) -> None:
        """Manually invalidate the cache.

        Use this when a new prompt version is created to force
        immediate reload on next access.
        """
        async with self._lock:
            if self._cached_prompt is not None:
                logger.debug(
                    "Cache invalidated (was version %d)",
                    self._cached_version,
                )
                self._cached_prompt = None
                self._cached_version = None
                self._cache_timestamp = None
            else:
                logger.debug("Cache invalidation called but cache was empty")

    def set_ttl(self, seconds: int) -> None:
        """Update cache TTL.

        Args:
            seconds: New TTL in seconds.
        """
        self._ttl_seconds = seconds
        logger.debug("Cache TTL updated to %d seconds", seconds)

    @property
    def is_cached(self) -> bool:
        """Check if prompt is currently cached and valid."""
        return self._is_cache_valid()

    @property
    def cached_version(self) -> int | None:
        """Get the currently cached prompt version (if any)."""
        return self._cached_version if self._is_cache_valid() else None


# Global cache instance
_prompt_cache = SystemPromptCache()


async def get_system_prompt() -> str:
    """Convenience function to get the current system prompt.

    Returns:
        The current system prompt text.
    """
    return await _prompt_cache.get_prompt()


async def invalidate_prompt_cache() -> None:
    """Convenience function to invalidate the prompt cache."""
    await _prompt_cache.invalidate()


def get_cache_instance() -> SystemPromptCache:
    """Get the global cache instance for advanced usage."""
    return _prompt_cache



"""Dialog components for skill management."""

import reflex as rx

import appkit_mantine as mn
from appkit_assistant.backend.database.models import Skill
from appkit_assistant.state.skill_admin_state import SkillAdminState
from appkit_ui.components.dialogs import delete_dialog

UPLOAD_ID = "skill_zip_upload"


def create_skill_modal() -> rx.Component:
    """Modal for uploading a new skill zip file."""
    return mn.modal(
        mn.stack(
            rx.upload(
                rx.cond(
                    rx.selected_files(UPLOAD_ID),
                    mn.stack(
                        mn.text(
                            rx.selected_files(UPLOAD_ID)[0],
                            size="sm",
                            fw="500",
                        ),
                        mn.text(
                            "Klicken zum Ändern",
                            size="xs",
                            c="dimmed",
                        ),
                        align="center",
                        gap="xs",
                    ),
                    mn.stack(
                        mn.text(
                            "ZIP-Datei hierher ziehen oder klicken",
                            size="sm",
                        ),
                        mn.text(
                            "Nur .zip Dateien",
                            size="xs",
                            c="dimmed",
                        ),
                        align="center",
                        gap="xs",
                    ),
                ),
                id=UPLOAD_ID,
                accept={
                    "application/zip": [".zip"],
                    "application/x-zip-compressed": [".zip"],
                },
                max_files=1,
                border="1px dashed var(--mantine-color-gray-5)",
                border_radius="var(--mantine-radius-md)",
                padding="2em",
                width="100%",
                cursor="pointer",
            ),
            mn.group(
                mn.button(
                    "Abbrechen",
                    variant="subtle",
                    on_click=SkillAdminState.close_create_modal,
                ),
                mn.button(
                    "Hochladen",
                    on_click=SkillAdminState.handle_upload(
                        rx.upload_files(upload_id=UPLOAD_ID)
                    ),
                    loading=SkillAdminState.uploading,
                    disabled=~rx.selected_files(UPLOAD_ID),
                ),
                w="100%",
                gap="9px",
                mt="12px",
                justify="flex-end",
            ),
            gap="12px",
            w="100%",
        ),
        title="Neuen Skill anlegen",
        opened=SkillAdminState.create_modal_open,
        on_close=SkillAdminState.close_create_modal,
        size="md",
        centered=True,
        overlay_props={"backgroundOpacity": 0.5, "blur": 4},
    )


def delete_skill_dialog(skill: Skill) -> rx.Component:
    """Delete confirmation dialog for a skill."""
    return delete_dialog(
        title="Skill löschen",
        content=skill.name,
        on_click=lambda: SkillAdminState.delete_skill(skill.id),
        icon_button=True,
        variant="subtle",
        color="red",
    )



"""Repository for MCP server data access operations."""

import logging
from datetime import UTC, datetime
from typing import Any

from sqlalchemy import or_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import defer

from appkit_assistant.backend.database.models import (
    AssistantAIModel,
    AssistantFileUpload,
    AssistantThread,
    MCPServer,
    Skill,
    SystemPrompt,
    UserPrompt,
    UserSkillSelection,
)
from appkit_commons.database.base_repository import BaseRepository
from appkit_user.authentication.backend.database import UserEntity

logger = logging.getLogger(__name__)


class MCPServerRepository(BaseRepository[MCPServer, AsyncSession]):
    """Repository class for MCP server database operations."""

    @property
    def model_class(self) -> type[MCPServer]:
        return MCPServer

    async def find_all_ordered_by_name(self, session: AsyncSession) -> list[MCPServer]:
        """Retrieve all MCP servers ordered by name."""
        stmt = select(MCPServer).order_by(MCPServer.name)
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_all_active_ordered_by_name(
        self, session: AsyncSession
    ) -> list[MCPServer]:
        """Retrieve all active MCP servers ordered by name."""
        stmt = (
            select(MCPServer)
            .where(MCPServer.active == True)  # noqa: E712
            .order_by(MCPServer.name)
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())


class SystemPromptRepository(BaseRepository[SystemPrompt, AsyncSession]):
    """Repository class for system prompt database operations.

    Implements append-only versioning with full CRUD capabilities.
    """

    @property
    def model_class(self) -> type[SystemPrompt]:
        return SystemPrompt

    async def find_all_ordered_by_version_desc(
        self, session: AsyncSession
    ) -> list[SystemPrompt]:
        """Retrieve all system prompt versions ordered by version descending."""
        stmt = select(SystemPrompt).order_by(SystemPrompt.version.desc())
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_latest(self, session: AsyncSession) -> SystemPrompt | None:
        """Retrieve the latest system prompt version."""
        stmt = select(SystemPrompt).order_by(SystemPrompt.version.desc()).limit(1)
        result = await session.execute(stmt)
        return result.scalars().first()

    async def create_next_version(
        self, session: AsyncSession, prompt: str, user_id: int
    ) -> SystemPrompt:
        """Neue System Prompt Version anlegen.

        Version ist fortlaufende Ganzzahl, beginnend bei 1.
        """
        stmt = select(SystemPrompt).order_by(SystemPrompt.version.desc()).limit(1)
        result = await session.execute(stmt)
        latest = result.scalars().first()
        next_version = (latest.version + 1) if latest else 1

        name = f"Version {next_version}"

        system_prompt = SystemPrompt(
            name=name,
            prompt=prompt,
            version=next_version,
            user_id=user_id,
            created_at=datetime.now(UTC),
        )
        session.add(system_prompt)
        await session.flush()
        await session.refresh(system_prompt)

        logger.debug(
            "Created system prompt version %s for user %s",
            next_version,
            user_id,
        )
        return system_prompt


class ThreadRepository(BaseRepository[AssistantThread, AsyncSession]):
    """Repository class for Thread database operations."""

    @property
    def model_class(self) -> type[AssistantThread]:
        return AssistantThread

    async def find_by_user(
        self, session: AsyncSession, user_id: int
    ) -> list[AssistantThread]:
        """Retrieve all threads for a user."""
        stmt = (
            select(AssistantThread)
            .where(AssistantThread.user_id == user_id)
            .order_by(AssistantThread.updated_at.desc())
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_by_thread_id(
        self, session: AsyncSession, thread_id: str
    ) -> AssistantThread | None:
        """Retrieve a thread by its thread_id."""
        stmt = select(AssistantThread).where(AssistantThread.thread_id == thread_id)
        result = await session.execute(stmt)
        return result.scalars().first()

    async def find_by_thread_id_and_user(
        self, session: AsyncSession, thread_id: str, user_id: int
    ) -> AssistantThread | None:
        """Retrieve a thread by thread_id and user_id."""
        stmt = select(AssistantThread).where(
            AssistantThread.thread_id == thread_id,
            AssistantThread.user_id == user_id,
        )
        result = await session.execute(stmt)
        return result.scalars().first()

    async def delete_by_thread_id_and_user(
        self, session: AsyncSession, thread_id: str, user_id: int
    ) -> bool:
        """Delete a thread by thread_id and user_id."""
        stmt = select(AssistantThread).where(
            AssistantThread.thread_id == thread_id,
            AssistantThread.user_id == user_id,
        )
        result = await session.execute(stmt)
        thread = result.scalars().first()
        if thread:
            await session.delete(thread)
            await session.flush()
            return True
        return False

    async def find_summaries_by_user(
        self, session: AsyncSession, user_id: int
    ) -> list[AssistantThread]:
        """Retrieve thread summaries (no messages) for a user."""
        stmt = (
            select(AssistantThread)
            .where(AssistantThread.user_id == user_id)
            .options(defer(AssistantThread.messages))
            .order_by(AssistantThread.updated_at.desc())
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_unique_vector_store_ids(self, session: AsyncSession) -> list[str]:
        """Get unique vector store IDs from all threads.

        Returns:
            List of unique vector store IDs (excluding None/empty).
        """
        stmt = select(AssistantThread.vector_store_id).distinct()
        result = await session.execute(stmt)
        return [row[0] for row in result.all() if row[0]]

    async def clear_vector_store_id(
        self, session: AsyncSession, vector_store_id: str
    ) -> int:
        """Clear vector_store_id from all threads referencing the given store.

        Args:
            vector_store_id: The vector store ID to clear from threads.

        Returns:
            Number of threads updated.
        """
        stmt = select(AssistantThread).where(
            AssistantThread.vector_store_id == vector_store_id
        )
        result = await session.execute(stmt)
        threads = list(result.scalars().all())

        for thread in threads:
            thread.vector_store_id = None
            session.add(thread)

        await session.flush()
        return len(threads)


class FileUploadRepository(BaseRepository[AssistantFileUpload, AsyncSession]):
    """Repository class for file upload database operations."""

    @property
    def model_class(self) -> type[AssistantFileUpload]:
        return AssistantFileUpload

    async def find_unique_vector_stores(
        self, session: AsyncSession
    ) -> list[tuple[str, str]]:
        """Get unique vector store IDs with names from all file uploads.

        Returns:
            List of tuples (vector_store_id, vector_store_name).
        """
        stmt = (
            select(
                AssistantFileUpload.vector_store_id,
                AssistantFileUpload.vector_store_name,
            )
            .distinct()
            .order_by(AssistantFileUpload.vector_store_id)
        )
        result = await session.execute(stmt)
        return [(row[0], row[1] or "") for row in result.all()]

    async def find_unique_vector_stores_by_ai_model(
        self, session: AsyncSession, ai_model: str
    ) -> list[tuple[str, str]]:
        """Get unique vector store IDs with names for a specific AI model.

        Args:
            ai_model: The model_id string to filter by.

        Returns:
            List of tuples (vector_store_id, vector_store_name).
        """
        stmt = (
            select(
                AssistantFileUpload.vector_store_id,
                AssistantFileUpload.vector_store_name,
            )
            .where(AssistantFileUpload.ai_model == ai_model)
            .distinct()
            .order_by(AssistantFileUpload.vector_store_id)
        )
        result = await session.execute(stmt)
        return [(row[0], row[1] or "") for row in result.all()]

    async def find_distinct_ai_models(self, session: AsyncSession) -> list[str]:
        """Get distinct ai_model values from all file uploads.

        Returns:
            List of unique ai_model strings (excluding empty).
        """
        stmt = (
            select(AssistantFileUpload.ai_model)
            .distinct()
            .order_by(AssistantFileUpload.ai_model)
        )
        result = await session.execute(stmt)
        return [row[0] for row in result.all() if row[0]]

    async def find_by_vector_store(
        self, session: AsyncSession, vector_store_id: str
    ) -> list[AssistantFileUpload]:
        """Get all files for a specific vector store."""
        stmt = (
            select(AssistantFileUpload)
            .where(AssistantFileUpload.vector_store_id == vector_store_id)
            .order_by(AssistantFileUpload.created_at.desc())
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_by_thread(
        self, session: AsyncSession, thread_id: int
    ) -> list[AssistantFileUpload]:
        """Get all files for a specific thread."""
        stmt = (
            select(AssistantFileUpload)
            .where(AssistantFileUpload.thread_id == thread_id)
            .order_by(AssistantFileUpload.created_at.desc())
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def delete_file(
        self, session: AsyncSession, file_id: int
    ) -> AssistantFileUpload | None:
        """Delete a file upload by ID and return the deleted record."""
        stmt = select(AssistantFileUpload).where(AssistantFileUpload.id == file_id)
        result = await session.execute(stmt)
        file_upload = result.scalars().first()
        if file_upload:
            await session.delete(file_upload)
            await session.flush()
            return file_upload
        return None

    async def delete_by_vector_store(
        self, session: AsyncSession, vector_store_id: str
    ) -> list[AssistantFileUpload]:
        """Delete all files for a vector store and return the deleted records."""
        stmt = select(AssistantFileUpload).where(
            AssistantFileUpload.vector_store_id == vector_store_id
        )
        result = await session.execute(stmt)
        files = list(result.scalars().all())
        for file_upload in files:
            await session.delete(file_upload)
        await session.flush()
        return files


class UserPromptRepository(BaseRepository[UserPrompt, AsyncSession]):
    """Repository for user prompts (single table design)."""

    @property
    def model_class(self) -> type[UserPrompt]:
        return UserPrompt

    async def find_latest_prompts_by_user(
        self, session: AsyncSession, user_id: int
    ) -> list[UserPrompt]:
        """Find latest versions of prompts owned by the user.

        Used for the sidebar list.
        """
        stmt = (
            select(UserPrompt)
            .where(
                UserPrompt.user_id == user_id,
                UserPrompt.is_latest == True,  # noqa: E712
            )
            .order_by(UserPrompt.handle)
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_latest_shared_prompts(
        self, session: AsyncSession, user_id: int
    ) -> list[dict[str, Any]]:
        """Find latest shared prompts (excluding own).

        Returns a list of dicts containing the prompt and the creator's name.
        """

        stmt = (
            select(UserPrompt, UserEntity.name)
            .join(UserEntity, UserPrompt.user_id == UserEntity.id)
            .where(
                UserPrompt.is_shared == True,  # noqa: E712
                UserPrompt.user_id != user_id,
                UserPrompt.is_latest == True,  # noqa: E712
            )
            .order_by(UserPrompt.handle)
        )

        result = await session.execute(stmt)
        prompts = []
        for prompt, creator_name in result:
            prompt_dict = prompt.model_dump()
            prompt_dict["creator_name"] = creator_name or "Unbekannt"
            # Ensure mcp_server_ids is included as a list
            prompt_dict["mcp_server_ids"] = list(prompt.mcp_server_ids)
            prompts.append(prompt_dict)

        return prompts

    async def find_all_versions(
        self, session: AsyncSession, user_id: int, handle: str
    ) -> list[UserPrompt]:
        """Find all versions of a specific prompt (by handle)."""
        stmt = (
            select(UserPrompt)
            .where(UserPrompt.user_id == user_id, UserPrompt.handle == handle)
            .order_by(UserPrompt.version.desc())
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_latest_by_handle(
        self, session: AsyncSession, user_id: int, handle: str
    ) -> UserPrompt | None:
        """Find the latest version of a prompt by handle."""
        stmt = select(UserPrompt).where(
            UserPrompt.user_id == user_id,
            UserPrompt.handle == handle,
            UserPrompt.is_latest == True,  # noqa: E712
        )
        result = await session.execute(stmt)
        return result.scalars().first()

    async def find_latest_accessible_by_handle(
        self, session: AsyncSession, user_id: int, handle: str
    ) -> UserPrompt | None:
        """Find the latest version of a prompt accessible to the user.

        Returns the prompt if it's owned by the user OR shared by another user.
        Prefers user's own prompt if both exist with the same handle.
        """
        stmt = (
            select(UserPrompt)
            .where(
                UserPrompt.handle == handle,
                UserPrompt.is_latest == True,  # noqa: E712
                or_(
                    UserPrompt.user_id == user_id,
                    UserPrompt.is_shared == True,  # noqa: E712
                ),
            )
            # Prefer user's own prompt over shared ones
            .order_by((UserPrompt.user_id == user_id).desc())
        )
        result = await session.execute(stmt)
        return result.scalars().first()

    async def validate_handle_unique(
        self, session: AsyncSession, user_id: int, handle: str
    ) -> bool:
        """Check if a handle is unique for the user.

        Returns True if unique (safe to use), False otherwise.
        """
        existing = await self.find_latest_by_handle(session, user_id, handle)
        return existing is None

    async def create_new_prompt(
        self,
        session: AsyncSession,
        user_id: int,
        handle: str,
        description: str,
        prompt_text: str,
        is_shared: bool = False,
        mcp_server_ids: list[int] | None = None,
    ) -> UserPrompt:
        """Create the first version of a new prompt."""
        prompt = UserPrompt(
            user_id=user_id,
            handle=handle,
            description=description,
            prompt_text=prompt_text,
            version=1,
            is_latest=True,
            is_shared=is_shared,
            mcp_server_ids=mcp_server_ids or [],
            created_at=datetime.now(UTC),
        )
        session.add(prompt)
        await session.flush()
        await session.refresh(prompt)
        return prompt

    async def create_next_version(
        self,
        session: AsyncSession,
        user_id: int,
        handle: str,
        description: str | None = None,
        prompt_text: str | None = None,
        is_shared: bool | None = None,
        mcp_server_ids: list[int] | None = None,
    ) -> UserPrompt:
        """Create a new version for an existing prompt.

        Updates the old latest version's is_latest to False.
        Copies over values from previous version if not provided.
        """
        # 1. Get current latest
        current_latest = await self.find_latest_by_handle(session, user_id, handle)
        if not current_latest:
            raise ValueError(
                f"Prompt with handle '{handle}' not found for user {user_id}"
            )

        # 2. Update metadata on old latest (optional, but 'is_latest' MUST change)
        current_latest.is_latest = False
        session.add(current_latest)

        # 3. Prepare new values (use new if provided, else copy from old)
        new_desc = (
            description if description is not None else current_latest.description
        )
        new_text = (
            prompt_text if prompt_text is not None else current_latest.prompt_text
        )
        new_shared = is_shared if is_shared is not None else current_latest.is_shared
        new_mcp_servers = (
            mcp_server_ids
            if mcp_server_ids is not None
            else list(current_latest.mcp_server_ids)
        )

        # 4. Create new version
        new_version = UserPrompt(
            user_id=user_id,
            handle=handle,
            description=new_desc,
            prompt_text=new_text,
            version=current_latest.version + 1,
            is_latest=True,
            is_shared=new_shared,
            mcp_server_ids=new_mcp_servers,
            created_at=datetime.now(UTC),
        )
        session.add(new_version)
        await session.flush()
        await session.refresh(new_version)

        return new_version

    async def delete_all_versions(
        self, session: AsyncSession, user_id: int, handle: str
    ) -> bool:
        """Delete all versions of a prompt (hard delete)."""
        stmt = select(UserPrompt).where(
            UserPrompt.user_id == user_id, UserPrompt.handle == handle
        )
        result = await session.execute(stmt)
        prompts = result.scalars().all()

        if not prompts:
            return False

        for p in prompts:
            await session.delete(p)

        await session.flush()
        return True

    async def update_handle(
        self, session: AsyncSession, user_id: int, old_handle: str, new_handle: str
    ) -> bool:
        """Update handle for all versions of a prompt.

        Returns True if successful, False if no prompts found.
        """
        stmt = select(UserPrompt).where(
            UserPrompt.user_id == user_id, UserPrompt.handle == old_handle
        )
        result = await session.execute(stmt)
        prompts = list(result.scalars().all())

        if not prompts:
            return False

        for p in prompts:
            p.handle = new_handle
            session.add(p)

        await session.flush()
        return True

    async def find_all_accessible_prompts_filtered(
        self, session: AsyncSession, user_id: int, filter_text: str = ""
    ) -> list[dict[str, Any]]:
        """Find all prompts accessible to the user, filtered by handle.

        Returns user's own prompts and shared prompts from others.
        Filters by handle.startswith(filter_text) if provided.

        Returns list of dicts with handle, description, prompt_text, mcp_server_ids,
        is_own.
        """
        # Query own prompts
        own_stmt = select(UserPrompt).where(
            UserPrompt.user_id == user_id,
            UserPrompt.is_latest == True,  # noqa: E712
        )
        if filter_text:
            own_stmt = own_stmt.where(UserPrompt.handle.startswith(filter_text))
        own_stmt = own_stmt.order_by(UserPrompt.handle)

        own_result = await session.execute(own_stmt)
        own_prompts = own_result.scalars().all()

        # Query shared prompts from others
        shared_stmt = select(UserPrompt).where(
            UserPrompt.is_shared == True,  # noqa: E712
            UserPrompt.user_id != user_id,
            UserPrompt.is_latest == True,  # noqa: E712
        )
        if filter_text:
            shared_stmt = shared_stmt.where(UserPrompt.handle.startswith(filter_text))
        shared_stmt = shared_stmt.order_by(UserPrompt.handle)

        shared_result = await session.execute(shared_stmt)
        shared_prompts = shared_result.scalars().all()

        # Combine and return as dicts
        result: list[dict[str, Any]] = [
            {
                "handle": p.handle,
                "description": p.description,
                "prompt_text": p.prompt_text,
                "mcp_server_ids": list(p.mcp_server_ids),
                "is_own": True,
            }
            for p in own_prompts
        ]

        result.extend(
            {
                "handle": p.handle,
                "description": p.description,
                "prompt_text": p.prompt_text,
                "mcp_server_ids": list(p.mcp_server_ids),
                "is_own": False,
            }
            for p in shared_prompts
        )

        return result


class SkillRepository(BaseRepository[Skill, AsyncSession]):
    """Repository for OpenAI skill database operations."""

    @property
    def model_class(self) -> type[Skill]:
        return Skill

    async def update_required_role(
        self, session: AsyncSession, skill_id: int, required_role: str | None
    ) -> bool:
        """Update the required role for a skill."""
        skill = await session.get(Skill, skill_id)
        if not skill:
            return False

        skill.required_role = required_role
        session.add(skill)
        await session.flush()
        await session.refresh(skill)
        return True

    async def find_all_ordered_by_name(self, session: AsyncSession) -> list[Skill]:
        """Retrieve all skills ordered by name."""
        stmt = select(Skill).order_by(Skill.name)
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_all_active_ordered_by_name(
        self, session: AsyncSession
    ) -> list[Skill]:
        """Retrieve all active skills ordered by name."""
        stmt = (
            select(Skill)
            .where(Skill.active == True)  # noqa: E712
            .order_by(Skill.name)
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_by_openai_id(
        self, session: AsyncSession, openai_id: str
    ) -> Skill | None:
        """Find a skill by its OpenAI ID."""
        stmt = select(Skill).where(Skill.openai_id == openai_id)
        result = await session.execute(stmt)
        return result.scalars().first()

    async def find_all_by_api_key_hash(
        self, session: AsyncSession, api_key_hash: str
    ) -> list[Skill]:
        """Retrieve all skills for a given API key hash, ordered by name."""
        stmt = (
            select(Skill).where(Skill.api_key_hash == api_key_hash).order_by(Skill.name)
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_all_active_by_api_key_hash(
        self, session: AsyncSession, api_key_hash: str
    ) -> list[Skill]:
        """Retrieve all active skills for a given API key hash."""
        stmt = (
            select(Skill)
            .where(
                Skill.api_key_hash == api_key_hash,
                Skill.active == True,  # noqa: E712
            )
            .order_by(Skill.name)
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())


class UserSkillRepository(BaseRepository[UserSkillSelection, AsyncSession]):
    """Repository for user skill selection operations."""

    @property
    def model_class(self) -> type[UserSkillSelection]:
        return UserSkillSelection

    async def find_by_user_id(
        self, session: AsyncSession, user_id: int
    ) -> list[UserSkillSelection]:
        """Retrieve all skill selections for a user."""
        stmt = select(UserSkillSelection).where(UserSkillSelection.user_id == user_id)
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def upsert(
        self,
        session: AsyncSession,
        user_id: int,
        skill_openai_id: str,
        enabled: bool,
    ) -> UserSkillSelection:
        """Insert or update a user skill selection."""
        stmt = select(UserSkillSelection).where(
            UserSkillSelection.user_id == user_id,
            UserSkillSelection.skill_openai_id == skill_openai_id,
        )
        result = await session.execute(stmt)
        existing = result.scalars().first()

        if existing:
            existing.enabled = enabled
            session.add(existing)
            await session.flush()
            await session.refresh(existing)
            return existing

        selection = UserSkillSelection(
            user_id=user_id,
            skill_openai_id=skill_openai_id,
            enabled=enabled,
        )
        session.add(selection)
        await session.flush()
        await session.refresh(selection)
        return selection

    async def delete_by_skill_openai_id(
        self, session: AsyncSession, skill_openai_id: str
    ) -> int:
        """Delete all user selections for a skill (cascade cleanup)."""
        stmt = select(UserSkillSelection).where(
            UserSkillSelection.skill_openai_id == skill_openai_id
        )
        result = await session.execute(stmt)
        selections = list(result.scalars().all())
        for s in selections:
            await session.delete(s)
        await session.flush()
        return len(selections)


class AIModelRepository(BaseRepository[AssistantAIModel, AsyncSession]):
    """Repository for AI model database operations."""

    @property
    def model_class(self) -> type[AssistantAIModel]:
        return AssistantAIModel

    async def find_all_ordered_by_text(
        self, session: AsyncSession
    ) -> list[AssistantAIModel]:
        """Retrieve all AI models ordered by display text."""
        stmt = select(AssistantAIModel).order_by(AssistantAIModel.text)
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_all_active_ordered_by_text(
        self, session: AsyncSession
    ) -> list[AssistantAIModel]:
        """Retrieve all active models ordered by text."""
        stmt = (
            select(AssistantAIModel)
            .where(AssistantAIModel.active == True)  # noqa: E712
            .order_by(AssistantAIModel.text)
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_by_model_id(
        self, session: AsyncSession, model_id: str
    ) -> AssistantAIModel | None:
        """Find a model by its unique model_id string."""
        stmt = select(AssistantAIModel).where(AssistantAIModel.model_id == model_id)
        result = await session.execute(stmt)
        return result.scalars().first()

    async def find_by_processor_type(
        self, session: AsyncSession, processor_type: str
    ) -> list[AssistantAIModel]:
        """Find all active models for a given processor type."""
        stmt = (
            select(AssistantAIModel)
            .where(
                AssistantAIModel.processor_type == processor_type,
                AssistantAIModel.active == True,  # noqa: E712
            )
            .order_by(AssistantAIModel.text)
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_all_skill_capable(
        self, session: AsyncSession
    ) -> list[AssistantAIModel]:
        """Retrieve all OpenAI models that support skills, ordered by text."""
        stmt = (
            select(AssistantAIModel)
            .where(
                AssistantAIModel.supports_skills == True,  # noqa: E712
                AssistantAIModel.processor_type == "openai",
            )
            .order_by(AssistantAIModel.text)
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def find_all_with_attachments(
        self, session: AsyncSession
    ) -> list[AssistantAIModel]:
        """Retrieve all active models that support file attachments."""
        stmt = (
            select(AssistantAIModel)
            .where(
                AssistantAIModel.supports_attachments == True,  # noqa: E712
                AssistantAIModel.active == True,  # noqa: E712
            )
            .order_by(AssistantAIModel.text)
        )
        result = await session.execute(stmt)
        return list(result.scalars().all())

    async def update_active(
        self, session: AsyncSession, model_id: int, active: bool
    ) -> AssistantAIModel | None:
        """Update the active flag for a model."""
        rec = await session.get(AssistantAIModel, model_id)
        if not rec:
            return None
        rec.active = active
        session.add(rec)
        await session.flush()
        await session.refresh(rec)
        return rec

    async def update_role(
        self, session: AsyncSession, model_id: int, requires_role: str | None
    ) -> AssistantAIModel | None:
        """Update the required_role for a model."""
        rec = await session.get(AssistantAIModel, model_id)
        if not rec:
            return None
        rec.requires_role = requires_role
        session.add(rec)
        await session.flush()
        await session.refresh(rec)
        return rec


# Export instances
mcp_server_repo = MCPServerRepository()
system_prompt_repo = SystemPromptRepository()
thread_repo = ThreadRepository()
file_upload_repo = FileUploadRepository()
user_prompt_repo = UserPromptRepository()
skill_repo = SkillRepository()
user_skill_repo = UserSkillRepository()
ai_model_repo = AIModelRepository()
